#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "graph.h"
#include "exp.h"
#include "hop.h"
#include "nodeid.h"
#include "pick.h"
#include "route.h"
#include "stat.h"
#include "stress.h"
#include "util.h"

/************************************************/
/* initialize output file names                 */
/************************************************/

void exp12_init_output_file_names(char *topology_fn, char *out_dir,
        char *hop_outfile1, char *hop_outfile2,
        char *rdp_outfile, char *stress_outfile1, char *stress_outfile2,
        int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
        int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
        int group_size, int num_clusters, int vnn) {

  char param[MAX_FILENAME_LEN];

  //  int vnn = atoi(getenv("REXEC_MY_VNN"));

  // initialize output file names
  sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_SIZE, group_size, num_clusters, vnn);
  sprintf(hop_outfile1, "%s.hop.sp", param);
  sprintf(hop_outfile2, "%s.hop.tapestry", param);
  sprintf(rdp_outfile, "%s.hop.rdp", param);
  sprintf(stress_outfile1, "%s.stress.unicast", param);
  sprintf(stress_outfile2, "%s.stress.tapestry", param);

  printf("%s\n", hop_outfile1); printf("%s\n", hop_outfile2);
  printf("%s\n", rdp_outfile);
  printf("%s\n", stress_outfile1); printf("%s\n", stress_outfile2);
}

/************************************************/
/************************************************/

void assign_cluster_nodeID(Graph *g, int TAPESTRY_NODEID_SIZE, 
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
		int cluster_size, int max_num_clusters, int num_clusters, 
		int *cluster_array) {
  int picked_cluster_array[max_num_clusters];
  int used_tid_array[TAPESTRY_NAMESPACE_SIZE];
  Vertex *v, *u;                                
  char *node_id = NULL;
  int i, j, k, len, stub_size, suffix, id, index;
  char vname[256], *ptr;

  if (g->n < TAPESTRY_NAMESPACE_SIZE) {
    printf("ERROR:  # of vertices in graph (%ld) < ", g->n);
    printf("TAPESTRY_NAMESPACE_SIZE (%d)\n", TAPESTRY_NAMESPACE_SIZE);
    exit(EXIT_FAILURE);
  }

  // clear nodeid field of every vertex in g to NULL
  clear_nodeid_field(g);
    
  for (i = 0; i < TAPESTRY_NAMESPACE_SIZE; i++) {
    used_tid_array[i] = 0;
  }
  for (i = 0; i < num_clusters; i++) {
    for (j = 0; j < cluster_size; j++) {
      used_tid_array[j * TAPESTRY_NAMESPACE_SIZE / cluster_size + i] = 1;
    }
  } 

  // pick num_clusters clusters randomly
  for (i = 0; i < max_num_clusters; i++) {
    picked_cluster_array[i] = 0;
  }
  for (i = 0; i < num_clusters; i++) {
    index = gb_unif_rand(max_num_clusters);
    while (picked_cluster_array[index] == 1) {
      index = gb_unif_rand(max_num_clusters);
    }
    picked_cluster_array[index] = 1;
  }

  i = 0;			// 0 - (max_num_clusters -1)
  j = 0;			// 0 - (num_stubs - 1)
  suffix = 0;			// 0 - (num_clusters - 1)
  v = g->vertices;
  while (i < max_num_clusters && v < g->vertices + g->n) {
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) {
      ptr = strchr(vname, '.');
      if (j == cluster_array[i]) {
        printf("i = %d, domain # = %d: %s - %s (%d)\n", i, j,
                (v-atoi(ptr+1)-1)->name, (v-1)->name, atoi(ptr+1)+1);
        if (picked_cluster_array[i] == 1) {
          printf("picked: ");
          for (k = 0; k < cluster_size; k++) {
            u = v - 1 - k;
            assert(u->picked_t == 1);
            id = k * (TAPESTRY_NAMESPACE_SIZE / cluster_size) + suffix;
            node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
            tapestry_nodeID_itoa(id, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
            u->nodeid = node_id;
            printf("%s (%s) ", u->name, u->nodeid);
            used_tid_array[id] = 1;
          }
          printf("\n");
          suffix++;
        } else {
          for (k = 0; k < cluster_size; k++) {
            u = v - 1 - k;
            assert(u->picked_t == 1);
            index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
            while (used_tid_array[index] == 1) {
              index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
            }
            node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
            tapestry_nodeID_itoa(index, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
            u->nodeid = node_id;
            printf("%s (%s) ", u->name, u->nodeid);
            used_tid_array[index] = 1;
          }
          printf("\n");
        }
        i++;
      }
      j++;
    }
    if (i < max_num_clusters) {
      strcpy(vname, v->name);
      ptr = strrchr(vname, '.');
      *ptr = '\0';
      len = strlen(vname);
      // recurse through a stub domain's nodes
      while ((v+1) < g->vertices+g->n && strncmp(vname,(v+1)->name,len)==0) {
        v++;
      }
      stub_size = atoi(strrchr(v->name, '.') + 1);
      if (j == cluster_array[i]) {
        printf("i = %d, domain # = %d: %s - %s (%d)\n", i, j,
                (v-stub_size)->name, v->name, stub_size+1);
        if (picked_cluster_array[i] == 1) {
          printf("picked: suffix = %d ", suffix);
          for (k = 0; k < cluster_size; k++) {
            u = v - k;
            assert(u->picked_t == 1);
            id = k * (TAPESTRY_NAMESPACE_SIZE / cluster_size) + suffix;
            node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
            tapestry_nodeID_itoa(id, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
            u->nodeid = node_id;
            printf("%s (%s) ", u->name, u->nodeid);
            used_tid_array[id] = 1;
          }
          printf("\n");
          suffix++;
        } else {
          for (k = 0; k < cluster_size; k++) {
            u = v - k;
            assert(u->picked_t == 1);
            index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
            while (used_tid_array[index] == 1) {
              index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
            }
            node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));      
            tapestry_nodeID_itoa(index, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
            u->nodeid = node_id;                
            printf("%s (%s) ", u->name, u->nodeid);
            used_tid_array[index] = 1;
          }
          printf("\n");
        }  
        i++;
      }                                 
      j++;
      v++;
    }   
  }             

  if (suffix != num_clusters) {
    printf("ERROR:  suffix %d != num_clusters %d\n", suffix, num_clusters);
    exit(1);
  }
  if (i != max_num_clusters) {
    printf("ERROR:  i %d != max_num_clusters %d\n", i, max_num_clusters);
    exit(1); 
  } 

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    if (v->picked_t == 1 && v->nodeid == NULL) {
      index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
      while (used_tid_array[index] == 1) {
        index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
      }
      node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
      tapestry_nodeID_itoa(index, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
      v->nodeid = node_id;
      used_tid_array[index] = 1;
    }
  }
}

/************************************************/
/* assign clusters of cluster_size tapestry     */
/* node IDs such as (**wxyz) to vertices of 	*/
/* the Graph pointed to by g                    */
/************************************************/

void assign_cluster_nodeID_old(Graph *g, int TAPESTRY_NODEID_SIZE, 
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
		int cluster_size, int num_clusters) {

  Vertex *v, *u;
  char *node_id = NULL;
  int i, len, stub_size, suffix, id, index;
  int used[TAPESTRY_NAMESPACE_SIZE];
  char vname[256], *ptr;

  if (g->n < TAPESTRY_NAMESPACE_SIZE) {
    printf("ERROR:  # of vertices in graph (%ld) < ", g->n);
    printf("TAPESTRY_NAMESPACE_SIZE (%d)\n", TAPESTRY_NAMESPACE_SIZE);
    exit(EXIT_FAILURE);
  }

  for (i = 0; i < TAPESTRY_NAMESPACE_SIZE; i++) {
    used[i] = 0;
  }

  v = g->vertices;
  suffix = 0;
  while (v < g->vertices + g->n && suffix < num_clusters) {
    // recurse through a transit domain's nodes
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) {
      ptr = strchr(vname, '.');
      stub_size = atoi(ptr+1);
      printf("%s - %s (%d)\n", (v-1-stub_size)->name, (v-1)->name,
							stub_size+1);
      // if a transit domain has >= cluster_size nodes,
      // assign a cluster of cluster_size node IDs
      if ((stub_size + 1) >= cluster_size) {
        for (i = 0; i < cluster_size; i++) {
          u = v - 1 - i;
          assert(u->picked_t == 1);
          id = i * (TAPESTRY_NAMESPACE_SIZE / cluster_size) + suffix;
          node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
          tapestry_nodeID_itoa(id, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
          u->nodeid = node_id;
          used[id] = 1; 
          printf("%s ", node_id);
        }
        printf("\n");
        suffix++;
      }
    }

    strcpy(vname, v->name);
    ptr = strrchr(vname, '.');
    *ptr = '\0';
    len = strlen(vname);
    // recurse through a stub domain's nodes
    while ((v+1) < g->vertices+g->n && strncmp(vname,(v+1)->name,len)==0) {
      v++;
    }
    stub_size = atoi(strrchr(v->name, '.') + 1);
    printf("%s - %s (%d)\n", (v-stub_size)->name, v->name, stub_size+1);
    // if a stub domain has >= cluster_size nodes,
    // assign a cluster of cluster_size node IDs
    if ((stub_size + 1) >= cluster_size) {
      for (i = 0; i < cluster_size; i++) {
        u = v - i;
        assert(u->picked_t == 1);
        id = i * (TAPESTRY_NAMESPACE_SIZE / cluster_size) + suffix;
        node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
        tapestry_nodeID_itoa(id, node_id, TAPESTRY_NODEID_SIZE,
						TAPESTRY_NAMESPACE_BASE);
        u->nodeid = node_id;
        used[id] = 1;
        printf("%s ", node_id);
      }
      printf("\n");
      suffix++;
    }
    v++;
  }

  if (suffix != num_clusters) {
    printf("ERROR:  suffix %d != num_clusters %d\n", suffix, num_clusters);
    exit(1);
  }

  // if # of clusters is less than TAPESTRY_NAMESPACE_SIZE / cluster_size,
  // assign the rest of node IDs randomly
  i = cluster_size * num_clusters;
  if (num_clusters < TAPESTRY_NAMESPACE_SIZE / cluster_size) {
    for (; v < g->vertices + g->n && i < TAPESTRY_NAMESPACE_SIZE; v++) {
      if (v->picked_t == 1) {
        index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
        while (used[index] == 1) {
          index = gb_unif_rand(TAPESTRY_NAMESPACE_SIZE);
        }
        used[index] = 1;
        node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
        tapestry_nodeID_itoa(index, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
        v->nodeid = node_id;
        i++;
      }
    }
  }

  if (i != TAPESTRY_NAMESPACE_SIZE) {
    printf("ERROR:  i %d != TAPESTRY_NAMESPACE_SIZE %d\n", i, 
						TAPESTRY_NAMESPACE_SIZE);
    exit(1);
  }
}

/************************************************/
/************************************************/

void pick_tapestry_clusters(Graph *g, int num_stubs,
		int TAPESTRY_NAMESPACE_SIZE,
		int cluster_size, int max_num_clusters, int **cluster_array) {
  
//  int max_num_clusters = TAPESTRY_NAMESPACE_SIZE / cluster_size; 
  int stub_array[num_stubs];
  int i, j, k, index, len;
  Vertex *v, *u;                    
  char vname[256], stub_size[256], *ptr;
    
  *cluster_array = malloc(sizeof(int) * max_num_clusters);

  // pick max_num_clusters domains randomly to each contain cluster_size
  // tapestry nodes in Graph g
  for (i = 0; i < num_stubs; i++) {
    stub_array[i] = 0;
  } 
  for (i = 0; i < max_num_clusters; i++) { 
    index = gb_unif_rand(num_stubs);
    while (stub_array[index] == 1) {
      index = gb_unif_rand(num_stubs);
    }
    stub_array[index] = 1;
    (*cluster_array)[i] = index;                   
    // printf("%d\n", (*cluster_array)[i]);
  } 
  bubblesort(*cluster_array, max_num_clusters);
  for (i = 0; i < max_num_clusters; i++) {
    printf("%d %d\n", i, (*cluster_array)[i]);
  }
  i = 0;
  j = 0;
  v = g->vertices;
  while (i < max_num_clusters && v < g->vertices + g->n) {
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) { 
      ptr = strchr(vname, '.');
      if (j == (*cluster_array)[i]) {
        if ((atoi(ptr+1)+1) < cluster_size) {
          printf("ERROR:  transit domain size %d < cluster size %d\n",
					atoi(ptr+1)+1, cluster_size);
          exit(1);
        }
        printf("i = %d, domain # = %d: %s - %s (%d)\n", i, j,
		(v-atoi(ptr+1)-1)->name, (v-1)->name, atoi(ptr+1)+1);
        for (k = 0; k < cluster_size; k++) {
          u = v - 1 - k;
          assert(u->picked_t != 1);
          u->picked_t = 1;
        }
        i++;                    
      }
      j++;
    } 
    if (i < max_num_clusters) {
      strcpy(vname, v->name);
      ptr = strrchr(vname, '.');
      *ptr = '\0';
      len = strlen(vname);
      while ((v+1) < g->vertices+g->n && strncmp(vname,(v+1)->name,len)==0) {
        v++;
      }
      ptr = strrchr(v->name, '.');              
      strcpy(stub_size, ptr+1);
      if (j == (*cluster_array)[i]) {
        if ((atoi(ptr+1)+1) < cluster_size) {
          printf("ERROR:  stub domain size %d < cluster size %d\n",
					atoi(ptr+1)+1, cluster_size);
          exit(1);
        }
        printf("i = %d, domain # = %d: %s - %s (%d)\n", i, j,
		(v-atoi(stub_size))->name, v->name, atoi(stub_size)+1);
        for (k = 0; k < cluster_size; k++) {
          u = v - k;
          assert(u->picked_t != 1);
          u->picked_t = 1;
        }
        i++;
      }
      j++;
      v++;
    } 
  } 
} 

/************************************************/
/* pick cluster_size transit/stub nodes in	*/
/* TAPESTRY_NAMESPACE_SIZE / cluster_size	*/
/* stub domains to be tapestry nodes		*/
/************************************************/

void pick_tapestry_clusters_old(Graph *g, int TAPESTRY_NAMESPACE_SIZE,
			int cluster_size) {
  int num_clusters = TAPESTRY_NAMESPACE_SIZE / cluster_size;
  Vertex *v, *u;
  char vname[256], *ptr;
  int suffix, j, len, stub_size;

  v = g->vertices;
  suffix = 0;
  while ((v < (g->vertices + g->n)) && (suffix < num_clusters)) {  
    // recurse through a transit domain's nodes
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) {
      ptr = strchr(vname, '.');
      stub_size = atoi(ptr+1);
      printf("%s - %s (%d)\n", (v-1-stub_size)->name, (v-1)->name,
							stub_size+1);
      // if a transit domain has >= cluster_size nodes,
      // assign a cluster of cluster_size node IDs
      if ((stub_size + 1) >= cluster_size) {
        for (j = 0; j < cluster_size; j++) {
          u = v - 1 - j;
          assert(u->picked_t != 1);
          u->picked_t = 1;
        }
        suffix++;
      }
    }

    strcpy(vname, v->name);
    ptr = strrchr(vname, '.');                  
    *ptr = '\0';
    len = strlen(vname);
    // recurse through a stub domain's nodes
    while (((v+1) < g->vertices+g->n) && strncmp(vname,(v+1)->name,len)==0) {
      v++;
    } 
    stub_size = atoi(strrchr(v->name, '.') + 1);
    printf("%s - %s (%d)\n", (v-stub_size)->name, v->name, stub_size+1);
    
    // if a stub domain has >= cluster_size nodes, pick a cluster of 
    // cluster_size nodes to be tapestry nodes
    if ((stub_size + 1) >= cluster_size) {
      for (j = 0; j < cluster_size; j++) {
        u = v - j;
        assert(u->picked_t != 1);
        u->picked_t = 1;
      }
      suffix++;
    }
    v++;
  }

  // if # of transit/stub domains with >= cluster_size nodes is 
  // less than num_clusters, ERROR! 
  if (suffix < num_clusters) {
    printf("ERROR:  # of transit/stub domains with >= %d ", cluster_size);
    printf("nodes is less than %d\n", num_clusters);
    exit(1);
  }
}

/************************************************/
/* initialize experiment                        */
/* 1.  assign tapestry node IDs to vertices     */
/* 2.  build tapestry routing table             */  
/************************************************/
  
void exp12_init(Graph *g, char *topology_fn, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
		int cluster_size, int max_num_clusters, int num_clusters,
		int *cluster_array) {

  // assign cluster tapestry nodeIDs to g's vertices
  assign_cluster_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE, cluster_size, max_num_clusters,
	num_clusters, cluster_array);

  // print_all_arcs(g);

  // build tapestry routing table
  build_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
  // print tapestry routing table
  print_tapestry_routing_table(g, TAPESTRY_NODEID_SIZE,
                                TAPESTRY_NAMESPACE_BASE);
}

/************************************************/
/* run experiment                               */
/* 1.  initialiez output file names             */
/* 2.  calc # of hops from sender to each       */
/*     receiver via shortest path routing       */
/* 3.  calc # of hops from sender to each       */ 
/*     receiver via tapestry routing            */
/* 4.  compute RDP                              */
/* 5.  calc stress in unicast routing           */ 
/* 6.  calc stress in tapestry routing          */ 
/************************************************/
  
void exp12_run(Graph *g, char *topology_fn, char *out_dir, 
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE,
                int group_size, int num_clusters, int vnn,
		Vertex *sender, Vertex **recvr_array) {
  
  char hop_outfile1[MAX_FILENAME_LEN];
  char hop_outfile2[MAX_FILENAME_LEN]; 
  char rdp_outfile[MAX_FILENAME_LEN];
  char stress_outfile1[MAX_FILENAME_LEN];
  char stress_outfile2[MAX_FILENAME_LEN];
  long *hop_array1, *hop_array2; 
  
  printf("\ntopology size = %ld, tapestry size = %d, group size = %d, ",
                g->n, TAPESTRY_NAMESPACE_SIZE, group_size);
  printf("number of clusters = %d\n", num_clusters);
 
  exp12_init_output_file_names(topology_fn, out_dir,
                hop_outfile1, hop_outfile2, rdp_outfile,
                stress_outfile1, stress_outfile2,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, num_clusters, vnn);

  // calc # of hops from the sender to each member via shortest path
  hop_calc_sp(g, sender, recvr_array, group_size - 1, hop_outfile1);

  // calc # of hops from the sender to each member via tapestry routing
  hop_calc_tapestry(g, sender, recvr_array, group_size - 1, hop_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);

  // compute RDP = tapestry routing delay / shortest path delay
  hop_array1 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_array2 = (long *) malloc(sizeof(long) * (group_size - 1));
  hop_read(hop_array1, group_size - 1, hop_outfile1);
  hop_read(hop_array2, group_size - 1, hop_outfile2);
  rdp_calc(hop_array1, hop_array2, group_size - 1, rdp_outfile);

  // compute stress of a physical link = # of identifical copies of
  // a packet carried by a physical link
  clear_stress_field(g);
  stress_calc_unicast(g, sender, recvr_array, group_size-1, stress_outfile1);
  clear_stress_field(g);
  stress_calc_tapestry(g, sender, recvr_array, group_size-1, stress_outfile2,
        TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
}

/************************************************/
/* experiment 12				*/
/* receiver node identifier clustering		*/
/************************************************/

void exp_12() {
  char topology_fn[MAX_FILENAME_LEN] = "ts50000.gb";
  int num_stubs = 1520;
  char out_dir[MAX_FILENAME_LEN] = "out12_amb";
  char gb_infile[MAX_FILENAME_LEN];
  int group_size = 256;
  int cluster_size = 16;
//  float fraction[8] = {0.0, 0.0625, 0.125, 0.25, 0.5, 0.75, 0.875, 1.0};
  float fraction[5] = {0.0, 0.25, 0.5, 0.75, 1.0};
  int max_num_clusters = 16;
  int num_clusters;
  int vnn = 9, i;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  Vertex *sender, **recvr_array, *v;
  int *cluster_array;

  TAPESTRY_NODEID_SIZE = 6; 
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;		// quad digit
  TAPESTRY_NAMESPACE_SIZE = 4096;

  sprintf(gb_infile, "../data/%s", topology_fn);
  g = restore_graph(gb_infile); // restores graph g from a sgb data file

  if (g == NULL)
    printf("ERROR: graph not read from topology data file.\n");
  reset_arc_len(g);             // reset arc lengths to 1

  clear_picked_t_field(g);      // clear the picked_t field of every vertex

  clear_picked_m_field(g);

  // pick TAPESTRY_NAMESPACE_SIZE tapestry nodes
  pick_tapestry_clusters(g, num_stubs, TAPESTRY_NAMESPACE_SIZE,
			cluster_size, max_num_clusters, &cluster_array);

  // pick a random sender
  sender = pick_1(g);

  // pick random multicast receivers from the tapestry network
  recvr_array = (Vertex **) malloc(sizeof(Vertex*) * (group_size - 1));
  pick_n(g, recvr_array, group_size - 1);

  // pick (TAPESTRY_NAMESPACE_SIZE - cluster_size * max_num_clusters) more
  // vertices randomly to be tapestry nodes in Graph g
  for (i=0; i<(TAPESTRY_NAMESPACE_SIZE-max_num_clusters*cluster_size); i++) {
    v = g->vertices + gb_unif_rand(g->n);
    while (v->picked_t == 1) {
      v = g->vertices + gb_unif_rand(g->n);
    }
    v->picked_t = 1;
  }

  for (i = 0; i < 5; i++) {
    num_clusters = (int) (fraction[i] * max_num_clusters);
    exp12_init(g, topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
		cluster_size, max_num_clusters, num_clusters, cluster_array);
    exp12_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, num_clusters, vnn, sender, recvr_array);
  }
}

/************************************************/
/* MAIN						*/
/************************************************/

int main() {

  gb_init_rand(time(NULL));     // initialize the random number generator

  // experiment for 12
  exp_12();

  return 0;
}
