#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAX_FILENAME_LEN 128

int main(int argc, char **argv) {
  char topology_fn[MAX_FILENAME_LEN];
  char hop_sp_fn[MAX_FILENAME_LEN];
  char hop_tapestry_fn[MAX_FILENAME_LEN];
  char hop_rdp_fn[MAX_FILENAME_LEN];
  FILE *hop_sp_fp, *hop_tapestry_fp, *hop_rdp_fp; 
  long *hop_sp_array, *hop_tapestry_array;
  int filesize, n, i;
  float *hop_rdp_array, rdp, max_rdp = FLT_MIN;
  struct stat statbuf;

  if (argc < 2)
    exit(EXIT_FAILURE);
  
  // initialize filenames
  strcpy(topology_fn, argv[1]);
  sprintf(hop_sp_fn, "%s.hop.sp", topology_fn);
  sprintf(hop_tapestry_fn, "%s.hop.tapestry", topology_fn);
  sprintf(hop_rdp_fn, "%s.hop.rdp", topology_fn);

  // use stat to determine file size
  stat(hop_sp_fn, &statbuf);
  filesize = statbuf.st_size;
  n = filesize / sizeof(long);

  // allocate memory for sp, tapestry, and rdp arrays
  hop_sp_array = malloc(sizeof(long) * n);
  hop_tapestry_array = malloc(sizeof(long) * n);
  hop_rdp_array = malloc(sizeof(float) * n);

  // open, read, and close sp, tapestry, and rdp files
  hop_sp_fp = fopen(hop_sp_fn, "r");
  hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
  hop_rdp_fp = fopen(hop_rdp_fn, "r");
  fread(hop_sp_array, sizeof(long), n, hop_sp_fp);
  fread(hop_tapestry_array, sizeof(float), n, hop_tapestry_fp);
  fread(hop_rdp_array, sizeof(float), n, hop_rdp_fp);
  fclose(hop_sp_fp);
  fclose(hop_tapestry_fp);
  fclose(hop_rdp_fp);

  // print values of sp, tapestry, and rdp arrays in columns
  for (i = 0; i < n; i++) {
    rdp = hop_rdp_array[i];
    printf("%ld %ld %0.2f\n", hop_sp_array[i], hop_tapestry_array[i], rdp);
    if (rdp > max_rdp)
      max_rdp = rdp;
  }
  // printf("tapestry worst case rdp = %0.2f\n", max_rdp);

  return 0;
}
