#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "graph.h"
#include "exp.h"
#include "nodeid.h"
#include "pick.h"

/************************************************/
/* pick num_failed_links links randomly and     */
/* remove them from the Graph pointed to by g   */
/************************************************/

void inject_link_faults(Graph *g, int num_failed_links) {
  int i;
  Vertex *v, *u;
  Arc *first_arc, *a, *pred_a;

  for (i = 0; i < num_failed_links; i++) {
    first_arc = NULL;
    while (first_arc == NULL) {
      v = g->vertices + gb_unif_rand(g->n);
      first_arc = v->arcs;
    }
    v->arcs = first_arc->next;
    u = first_arc->tip;
    pred_a = NULL;
    for (a = u->arcs; a; a = a->next) {
      if (a->tip == v) {
        if (a == u->arcs) {
          u->arcs = a->next;
        } else {
          pred_a->next = a->next;
        }
        break;
      }
      pred_a = a;
    }
  }
}

/************************************************/
/* pick num_failed_nodes nodes (not in the      */
/* multicast group) randomly and remove them    */
/* from the Graph pointed to by g               */
/************************************************/

Graph *inject_node_faults(Graph *g, Vertex **new_sender,
                Vertex **new_recvr_array, int group_size,
                int num_failed_nodes,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
  Graph *new_graph;
  Vertex *v, *u, *w;
  int i, j, num_before;
  int routing_table_size = TAPESTRY_NODEID_SIZE * TAPESTRY_NAMESPACE_BASE *
                                MAX_NUM_NBRS;
  Vertex **v_routing_table, **u_routing_table;
  Vertex *remv_array[num_failed_nodes];

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->ind = 1;
  }
  for (i = 0; i < num_failed_nodes; i++) {
    v = g->vertices + gb_unif_rand(g->n);
    while (v->ind == 0 || v->picked_m == 1) {
      v = g->vertices + gb_unif_rand(g->n);
    }
    v->ind = 0;
    remv_array[i] = v;
    printf("remv_array[%d] = %s\n", i, v->name);
    fflush(stdout);
  }
  new_graph = induced(g, NULL, 0, 0, 0);

  reset_arc_len(new_graph);             // reset arc lengths to 1
  clear_picked_t_field(new_graph);      // clear the picked_t field
  clear_picked_m_field(new_graph);      // clear the picked_m field

  for (v = g->vertices, u = new_graph->vertices; v < g->vertices + g->n &&
                        u < new_graph->vertices + new_graph->n; v++, u++) {
    while (v->ind == 0)
      v++;
    u->picked_t = v->picked_t;
    u->picked_m = v->picked_m;
    u->nodeid = v->nodeid;

    if (u->picked_t == 1) {
      v_routing_table = (Vertex **) v->rtable;
      u_routing_table = (Vertex **)malloc(sizeof(Vertex*)*routing_table_size);
      for (i = 0; i < routing_table_size; i++) {
        w = v_routing_table[i];
        num_before = 0;
        for (j = 0; j < num_failed_nodes; j++) {
          if (remv_array[j] < w)
            num_before++;
          else if (remv_array[j] == w)
            break;
        }
        if (remv_array[j] == w)
          u_routing_table[i] = NULL;
        else
          u_routing_table[i] = new_graph->vertices +
                                (w - g->vertices - num_before);
      }
      u->rtable = (long) u_routing_table;
    }
  }

  // update new_sender to point to vertex in new_graph
  num_before = 0;
  for (j = 0; j < num_failed_nodes; j++) {
    if (remv_array[j] < *new_sender)
      num_before++;
    else if (remv_array[j] == *new_sender)
      break;
  }
  if (remv_array[j] == *new_sender) {
    printf("ERROR:  inject_node_faults removed sender!\n"); fflush(stdout);
    exit(1);
  }
  else
    *new_sender = new_graph->vertices +
                                (*new_sender - g->vertices - num_before);

  // update new_recvr_array to point to vertices in new_graph
  for (i = 0; i < group_size - 1; i++) {
    num_before = 0;
    for (j = 0; j < num_failed_nodes; j++) {
      if (remv_array[j] < new_recvr_array[i])
        num_before++;
      else if (remv_array[j] == new_recvr_array[i])
        break;
    }
    if (remv_array[j] == new_recvr_array[i]) {
      printf("ERROR:  inject_node_faults removed receiver %d!\n", i);
      fflush(stdout);
      exit(1);
    }
    else
      new_recvr_array[i] = new_graph->vertices +
                        (new_recvr_array[i] - g->vertices - num_before);
  }
  return new_graph;
}
