#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "exp.h"
#include "graph.h"
#include "pick.h"
#include "nodeid.h"
#include "route.h"

/************************************************/
/* experiment 3 & 8				*/
/* varying multicast group size			*/
/************************************************/

void exp_3_8() {
  char *topology_fn_array[9] = {"ts5000.gb",
	"ts10000.gb", "ts20000.gb", "ts30000.gb", "ts40000.gb", "ts50000.gb", 
	"ti5000.gb", "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out3_8";
  int i, group_size;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 6;		// 6
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;	// 4096

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 1; i < 5; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      for (group_size = 16; group_size <= 4096; group_size *= 2) {
        exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
		group_size, vnn);
      }
    }
  }
}

/************************************************/
/* experiment 5 & 10                            */
/* varying tapestry namespace size		*/
/************************************************/

void exp_5_10() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out5_10";
  int i, group_size = 64;
  int vnn;
  
  int TAPESTRY_NODEID_SIZE;		// 3, 4, 5, 6 
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE;		// 64, 256, 1024, 4096 

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 4; i++) {
      for (TAPESTRY_NODEID_SIZE = 3; TAPESTRY_NODEID_SIZE <= 6; 
						TAPESTRY_NODEID_SIZE++) {
        TAPESTRY_NAMESPACE_SIZE = pow(TAPESTRY_NAMESPACE_BASE, 
					TAPESTRY_NODEID_SIZE);
        g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
        exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
      }
    }
  }
}

/************************************************/
/* experiment 4 & 9				*/
/* varying topology size			*/
/************************************************/

void exp_4_9() {
  char *topology_fn_array[14] = {"ts100.gb", "ts500.gb", "ts1000.gb",
		"ts2000.gb", "ts3000.gb", "ts4000.gb", "ts5000.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out4_9";
  int i, group_size = 64;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 3;
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 64;

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 7; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
    }
  }
}

/************************************************/
/* experiment 11				*/
/* varying tapestry namespace base (fanout)	*/
/************************************************/

void exp_11() {
  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  int group_size = 64;
  int vnn;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE;
  Graph *g;

//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  for (vnn = 5; vnn < 10; vnn++) {
    TAPESTRY_NODEID_SIZE = 3; 
    TAPESTRY_DIGIT_SIZE = 4;
    TAPESTRY_NAMESPACE_BASE = 16;	// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 4; 
    TAPESTRY_DIGIT_SIZE = 3;
    TAPESTRY_NAMESPACE_BASE = 8;		// octal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 6; 
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;		// quad digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 12; 
    TAPESTRY_DIGIT_SIZE = 1;
    TAPESTRY_NAMESPACE_BASE = 2;		// binary digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
  }
}

Vertex * nextHop(Vertex * currV, Vertex * destV, int currStep,
		 int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_DIGIT_SIZE,
		 int TAPESTRY_NODEID_SIZE, int nbr) {
  Vertex * nextV;
  int mask, digit, r_id;
  Vertex ** routing_table;

  routing_table = (Vertex **) currV->rtable;

  mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * currStep);
  tapestry_nodeID_atoi(destV->nodeid, &r_id,
		       TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
  digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * currStep);
  
  nextV = routing_table[TAPESTRY_NAMESPACE_BASE * currStep * MAX_NUM_NBRS +
		       digit * MAX_NUM_NBRS + nbr];
  return nextV;
}

void convergence() {
  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  int i, j, k, l, z, RUNS=100, PATHS=10;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE;
  int TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE;
  char gb_infile[MAX_FILENAME_LEN];
  Vertex *client, *root, *v1;

  Graph *g;

  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;		// octal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
  
  sprintf(gb_infile, "../data/%s", topology_fn);
  g = restore_graph(gb_infile);

  if (g == NULL)
    printf("ERROR: graph not read from topology data file.\n");
  reset_arc_len(g);
  
  clear_picked_t_field(g);
  assign_tapestry_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
			 TAPESTRY_NAMESPACE_SIZE);
  build_tapestry_routing_table_maxconn(g, TAPESTRY_NODEID_SIZE,
				       TAPESTRY_NAMESPACE_BASE);
  clear_picked_m_field(g);

  printf("TrueHopDist, StartHop, Hops2Conv\n");
  for (z=0; z<PATHS; z++) {
    clear_picked_m_field(g);
    root = pick_1(g);

    for (i=0; i< RUNS; i++) {
      client = pick_1(g);
      v1 = client;
      for (j=0; j < TAPESTRY_NODEID_SIZE; j++) {
	v1 = nextHop(v1, root, j, TAPESTRY_NAMESPACE_BASE,
		     TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	v1->picked_m = 1;
      }
      //go across the primary path to root and mark each node

      v1 = client;
      for (k=0; k < TAPESTRY_NODEID_SIZE-1; k++) {
	//k iterates, represents # of hops before branching off
	for (j=0; j < k; j++) {
	  v1 = nextHop(v1, root, j, TAPESTRY_NAMESPACE_BASE,
		       TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	  v1->picked_m = 1;
	}
	v1 = nextHop(v1, root, j, TAPESTRY_NAMESPACE_BASE,
		     TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 2);
//	v2 = nextHop(v1, root, j, TAPESTRY_NAMESPACE_BASE,
//		     TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 2);
	j++;
//	m = j;
	for(l=1; j < TAPESTRY_NODEID_SIZE; j++,l++) {
	  v1 = nextHop(v1, root, j, TAPESTRY_NAMESPACE_BASE,
		       TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	  if (v1->picked_m == 1) break;
	}
	printf("%ld %d %d\n", dijkstra(client, root, g, NULL),  k, l);
//	for(l=1; m < TAPESTRY_NODEID_SIZE; m++,l++) {
//	  v2 = nextHop(v2, root, m, TAPESTRY_NAMESPACE_BASE,
//		       TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
//	  if (v2->picked_m == 1) break;
//	}
//	printf(" %d\n", l);
      }
    }

  }
}

void fault_tap_mip(char * datafile) {
  char topology_fn[MAX_FILENAME_LEN] = "ts5000-3.gb";
  int i, j, k, m, mask2, x;
  int l, z, RUNS=10, ROOTS=5, MAXEXPERIMENTS=5;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE;
  int TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE;
  char gb_infile[MAX_FILENAME_LEN];
  Vertex *mh, *ch, *v1, *v2, *v3, *r, *intersect;
  Vertex **rootids;
  int MAXMN = 5;
  float success[ROOTS];
  float distance, f;
  float prob;
  float a, b;
  float sum;

  Graph *g;

  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;		// octal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
  
  rootids = (Vertex **) malloc (sizeof(Vertex *) * ROOTS);

  if (datafile == NULL) {
    printf("STARTING MOBILE TEST with %s.\n", topology_fn);
    sprintf(gb_infile, "../data/%s", topology_fn);
  } else {
    printf("STARTING MOBILE TEST with %s.\n", datafile);
    sprintf(gb_infile, "../data/%s", datafile);
  }

  fflush(stdout);
  g = restore_graph(gb_infile);
  if (g == NULL)
    printf("ERROR: graph not read from topology data file.\n");
  reset_arc_len(g);
  
  for (k=0; k< MAXEXPERIMENTS; k++) {
    clear_picked_t_field(g);
    
    assign_tapestry_nodeID(g, TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_NAMESPACE_SIZE);
    build_tapestry_routing_table_maxconn(g, TAPESTRY_NODEID_SIZE,
					 TAPESTRY_NAMESPACE_BASE);
    for (prob = 0.99; prob > 0.89; prob = prob-0.01) {
      for (z=0; z<MAXMN; z++) { //going thru diff choices of mobile node points
	clear_picked_m_field(g);
	mh = pick_0(g);
	
	for (j=0; j<ROOTS; j++) {
	  //let's publish the same mobilenode on multiple root names
	  rootids[j] = pick_0(g);
	  mask2 = 1 << j;   // used to mark nodes for a certain object
	  mh->picked_m = mh->picked_m |mask2;
	  rootids[j]->picked_m = rootids[j]->picked_m |mask2;
	  
	  r = rootids[j];
	  v1 = mh;
	  for (i = 0; i<TAPESTRY_NODEID_SIZE; i++) {
	    v2 = nextHop(v1, r, i, TAPESTRY_NAMESPACE_BASE,
			 TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	    v2->picked_m = v2->picked_m | mask2;
	    v1 = v2;
	  }
	}  // done publishing from mobile node to multiple roots
	
	for (l=0; l<RUNS; l++) {      //iterate on runs
	  ch = pick_0(g);
	  dijkstra(ch, mh, g, NULL);
	  printf(" k%d mn%d: %f: %d: ", k, z, prob, (int) mh->dist);
	  
	  for (j=0; j< ROOTS; j++) {  //iterate on roots
	    distance =1;
	    v1 = ch;
	    r = rootids[j];
	    for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	      v2 = nextHop(v1, r, i, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 2);
	      dijkstra(v1, v2, g, NULL);
	      f = pow(prob, v2->dist);
	      
	      v2 = nextHop(v1, r, i, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 1);
	      dijkstra(v1, v2, g, NULL);
	      f = 1 - (1- pow(prob, v2->dist))*(1 - f);
	      
	      v2 = nextHop(v1, r, i, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	      dijkstra(v1, v2, g, NULL);
	      f = 1 - (1- pow(prob, v2->dist))*(1 - f);
	      //might need to differentiate between diff roots here??
	      if (v2->picked_m != 0) {
		distance = distance * pow(prob, v2->dist);
		break;
	      }
	      distance = distance * f;
	      v1 = v2;
	    }
	    //done with finding intersect, using v2 to store intersect point
	    intersect = v2;

	    v1 = mh;

	    for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	      v2 = nextHop(v1, r, i, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	      dijkstra(v1, v2, g, NULL);
	      a = pow(prob, v2->dist);

	      if (v2 == intersect) {
		//we're back at the intersection
		distance = distance * a;
		break;
	      }
	      v3 = nextHop(v2, r, i, TAPESTRY_NAMESPACE_BASE,
			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE, 0);
	      dijkstra(v1, v3, g, NULL);
	      b = pow(prob, v3->dist);
	      v1 = v2;
	      distance = distance *(1- (1-a)*(1-b));
	    }

	    printf("%f ", distance);
	  } //end for (on ROOTS)
	  
	  //begin mobileIP section

	  sum =0;
	  for (i=0; i<100; i++) {
	    v1 = pick_0(g);
	    dijkstra(ch, v1, g, NULL);
	    distance = v1->dist;
	    dijkstra(v1, mh, g, NULL);
	    distance += mh->dist;
	    if (mh->dist < 9) {
	      i--;
	    } else {
	      sum += pow(prob, distance);
	    }
	  }
	  printf(" MobileIP: %f\n", sum/100);
	} //end for (on RUNS)
      } //end for (on choice of mobile nodes)
    } //end for MAXEXPERIMENTS
  }
}	  
      

/************************************************/
/* MAIN						*/
/************************************************/

int main(int argc, char *argv[]) {

  gb_init_rand(time(NULL));     // initialize the random number generator

  // experiment for 3 & 8
  // exp_3_8();

  // experiment for 5 & 10
  // exp_5_10();

  // experiment for 4 & 9
  // exp_4_9();

  // experiment for 11
  //  exp_11();

  //  convergence();

  //  if (argc == 1)
  //    fault_tap_mip("ts5000-3.gb");
  //  else 
  fault_tap_mip(argv[1]);
  return 0;
}
