#include <assert.h>
#include <string.h>
#include <math.h>
#include <gb_graph.h>
#include <gb_flip.h>

#include "graph.h"
#include "nodeid.h"
#include "util.h"

/************************************************/
/* clear the nodeid field of each vertex of	*/
/* the Graph pointed to by g			*/
/************************************************/

void clear_nodeid_field(Graph *g) {
  Vertex *v;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->nodeid = NULL;
  }
}

/************************************************/
/* clear the picked_t field of each vertex of   */
/* the Graph pointed to by g                    */
/************************************************/

void clear_picked_t_field(Graph *g) {
  Vertex *v;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->picked_t = 0;
  }
}

/************************************************/
/* convert a tapestry nodeID in decimal value   */
/* to its hexadecimal/octal/quad/binary form    */
/************************************************/

void tapestry_nodeID_itoa(int id, char *node_id,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {
    
  char temp_id[TAPESTRY_NODEID_SIZE + 1];
  int num_of_zeros, i = 0;
  
  if (TAPESTRY_NAMESPACE_BASE == 2)             // binary
    binary_itoa(id, temp_id, TAPESTRY_NODEID_SIZE);
  else if (TAPESTRY_NAMESPACE_BASE == 4)        // quad
    quad_itoa(id, temp_id, TAPESTRY_NODEID_SIZE); 
  else if (TAPESTRY_NAMESPACE_BASE == 8)        // octal
    sprintf(temp_id, "%0o", id);
  else if (TAPESTRY_NAMESPACE_BASE == 16)       // hexadecimal
    sprintf(temp_id, "%0x", id);
  else {
    printf("ERROR:  TAPESTRY_NAMESPACE_BASE (%d) is not yet supported",
                                        TAPESTRY_NAMESPACE_BASE);
    exit(EXIT_FAILURE);
  }

  num_of_zeros = TAPESTRY_NODEID_SIZE - strlen(temp_id);
  for (i = 0; i < num_of_zeros; i++) {
    node_id[i] = '0';
  }
  node_id[i] = '\0';
  strcat(node_id, temp_id);

  assert(strlen(node_id) == TAPESTRY_NODEID_SIZE);
}

/************************************************/
/* convert a tapestry nodeID in                 */
/* hexadecimal/octal/quad/binary form to its    */
/* decimal value                                */
/************************************************/

void tapestry_nodeID_atoi(char *node_id, int *id,
                int TAPESTRY_NODEID_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int temp_id = 0, i;
  char c;

  for (i = 0; i < TAPESTRY_NODEID_SIZE; i++) {
    c = node_id[i];
    temp_id += hoqb_atoi(c) * pow(TAPESTRY_NAMESPACE_BASE,
                                TAPESTRY_NODEID_SIZE - 1 - i);
  }
  *id = temp_id;
}

/************************************************/
/* randomly assign tapestry node IDs to         */
/* vertices of the Graph pointed to by g        */
/************************************************/

void assign_tapestry_nodeID(Graph *g, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_BASE, int TAPESTRY_NAMESPACE_SIZE) {

  Vertex *v;
  char *node_id = NULL;
  int i;

  if (g->n < TAPESTRY_NAMESPACE_SIZE) {
    printf("ERROR:  # of vertices in graph (%ld) < ", g->n);
    printf("TAPESTRY_NAMESPACE_SIZE (%d)\n", TAPESTRY_NAMESPACE_SIZE);
    exit(EXIT_FAILURE);
  }

  for (i = 0; i < TAPESTRY_NAMESPACE_SIZE; i++) {
    node_id = malloc((TAPESTRY_NODEID_SIZE + 1) * sizeof(char));
    tapestry_nodeID_itoa(i, node_id, TAPESTRY_NODEID_SIZE,
                                                TAPESTRY_NAMESPACE_BASE);
    // pick a random vertex
    v = g->vertices + gb_unif_rand(g->n);
    while (v->picked_t == 1) {
      v = g->vertices + gb_unif_rand(g->n);
    }
    v->nodeid = node_id;
    v->picked_t = 1;
  }
}
