#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gb_graph.h>
#include <gb_dijk.h>

#include "exp.h"
#include "read.h"
#include "stat.h"

/************************************************/
/* process results from experiment 3 & 8	*/
/* varying multicast group size			*/
/************************************************/

void proc_3_8() {
  char *topology_fn_array[9] = {"ts5000.gb",
	"ts10000.gb", "ts20000.gb", "ts30000.gb", "ts40000.gb", "ts50000.gb",
	"ti5000.gb", "AS010200.gb", "mb022399.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out3_8";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN], plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  float *stress_u_array=(float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_t_array=(float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_r_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *rdp_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, group_size;

  int TAPESTRY_NODEID_SIZE = 6;		// 6
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  // int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;	// 4096

  for (i = 5; i < 6; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp3_8.%s.hop.out", plot_dir, topology_fn);
    sprintf(plot_file2, "%s/exp3_8.%s.stress.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
    plot_fp2 = fopen(plot_file2, "w");

    for (group_size = 16; group_size <= 4096; group_size *= 2) {
      read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
      mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
      std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%d %0.2f %0.2f\n", group_size, mean, std);
      read_stress(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size,
		stress_u_array, stress_t_array, stress_r_array);
      mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%d %0.2f %0.2f ", group_size, mean, std);
      mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
      mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);
    }
    fclose(plot_fp1);
    fclose(plot_fp2);
  }
}

/************************************************/
/* process results from experiment 5 & 10       */
/* varying tapestry namespace size		*/
/************************************************/

void proc_5_10() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out5_10";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN], plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  float *stress_u_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_t_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_r_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *rdp_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, group_size = 64;
  int TAPESTRY_NODEID_SIZE;		// 3, 4, 5, 6 
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE;		// 64, 256, 1024, 4096 

  for (i = 0; i < 4; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp5_10.%s.hop.out", plot_dir, topology_fn);
    sprintf(plot_file2, "%s/exp5_10.%s.stress.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
    plot_fp2 = fopen(plot_file2, "w");

    for (TAPESTRY_NODEID_SIZE = 3; TAPESTRY_NODEID_SIZE <= 6; 
						TAPESTRY_NODEID_SIZE++) {
      TAPESTRY_NAMESPACE_SIZE = pow(TAPESTRY_NAMESPACE_BASE, 
					TAPESTRY_NODEID_SIZE);
      read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
      mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
      std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%d %0.2f %0.2f\n",TAPESTRY_NAMESPACE_SIZE,mean,std);
      read_stress(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size,
		stress_u_array, stress_t_array, stress_r_array);
      mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%d %0.2f %0.2f ",TAPESTRY_NAMESPACE_SIZE,mean,std);
      mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%0.2f %0.2f ",mean,std);
      mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
      std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, "%0.2f %0.2f\n",mean,std);
    }
  }
}

/************************************************/
/* process results from experiment 4 & 9	*/
/* varying topology size			*/
/************************************************/

void proc_4_9() {
  char *topology_fn_array[7] = {"ts100.gb", "ts500.gb", "ts1000.gb",
		"ts2000.gb", "ts3000.gb", "ts4000.gb", "ts5000.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out4_9";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN], plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  float *stress_u_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_t_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_r_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *rdp_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, group_size = 64, topology_size;

  int TAPESTRY_NODEID_SIZE = 3;
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  // int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 64;

  sprintf(plot_file1, "%s/exp4_9.hop.out", plot_dir);
  sprintf(plot_file2, "%s/exp4_9.stress.out", plot_dir);
  plot_fp1 = fopen(plot_file1, "w");
  plot_fp2 = fopen(plot_file2, "w");

  for (i = 0; i < 7; i++) {
    topology_fn = topology_fn_array[i];
    sscanf(topology_fn, "ts%d.gb", &topology_size);

    read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", topology_size, mean, std);

    read_stress(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, 
		stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", topology_size, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);
  }

  fclose(plot_fp1);
  fclose(plot_fp2);
}

/************************************************/
/* process results from experiment 11		*/
/* varying tapestry namespace base (fanout)	*/
/************************************************/

void proc_11() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN], plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  int group_size = 64;
  int i;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE;
  float *stress_u_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_t_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_r_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *rdp_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;

  for (i = 0; i < 4; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp11.%s.hop.out", plot_dir, topology_fn);
    sprintf(plot_file2, "%s/exp11.%s.stress.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
    plot_fp2 = fopen(plot_file2, "w");

    TAPESTRY_NODEID_SIZE = 12; 
    TAPESTRY_DIGIT_SIZE = 1;
    TAPESTRY_NAMESPACE_BASE = 2;		// binary digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    read_delay(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", TAPESTRY_NAMESPACE_BASE, mean, std);
    read_stress(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_SIZE, group_size,
                stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", TAPESTRY_NAMESPACE_BASE, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);

    TAPESTRY_NODEID_SIZE = 6; 
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;		// quad digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    read_delay(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", TAPESTRY_NAMESPACE_BASE, mean, std);
    read_stress(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_SIZE, group_size,
                stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", TAPESTRY_NAMESPACE_BASE, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);

    TAPESTRY_NODEID_SIZE = 4; 
    TAPESTRY_DIGIT_SIZE = 3;
    TAPESTRY_NAMESPACE_BASE = 8;		// octal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    read_delay(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", TAPESTRY_NAMESPACE_BASE, mean, std);
    read_stress(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_SIZE, group_size,
                stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", TAPESTRY_NAMESPACE_BASE, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);

    TAPESTRY_NODEID_SIZE = 3; 
    TAPESTRY_DIGIT_SIZE = 4;
    TAPESTRY_NAMESPACE_BASE = 16;		// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    read_delay(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", TAPESTRY_NAMESPACE_BASE, mean, std);
    read_stress(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_SIZE, group_size,
                stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", TAPESTRY_NAMESPACE_BASE, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);

    fclose(plot_fp1);
    fclose(plot_fp2);
  }
}

/************************************************/
/* process results from experiment 12	        */
/* varying number of clusters                   */
/************************************************/

void proc_12() {
  char topology_fn[MAX_FILENAME_LEN] = {"ts50000.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out12_amb";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN], plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  float *stress_u_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_t_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *stress_r_array = (float *) malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float *rdp_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i;
  int group_size = 256;
//  int cluster_size = 16;
  int max_num_clusters = 16;
//  float fraction[8] = {0.0, 0.0625, 0.125, 0.25, 0.5, 0.75, 0.875, 1.0};
  float fraction[5] = {0.0, 0.25, 0.5, 0.75, 1.0};
  int num_clusters;

  int TAPESTRY_NODEID_SIZE = 6;
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
//  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;

  sprintf(plot_file1, "%s/exp12_amb.%s.hop.out", plot_dir, topology_fn);
  sprintf(plot_file2, "%s/exp12_amb.%s.stress.out", plot_dir, topology_fn);
  plot_fp1 = fopen(plot_file1, "w");
  plot_fp2 = fopen(plot_file2, "w");

  for (i = 0; i < 5; i++) {
    num_clusters = (int) (fraction[i] * max_num_clusters);
    exp12_read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, num_clusters, rdp_array);
    mean = calc_mean(rdp_array, MAX_NUM_EXPERIMENT);
    std = calc_std(rdp_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp1, "%d %0.2f %0.2f\n", num_clusters, mean, std);

    exp12_read_stress(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, num_clusters,
                stress_u_array, stress_t_array, stress_r_array);
    mean = calc_mean(stress_u_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_u_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%d %0.2f %0.2f ", num_clusters, mean, std);
    mean = calc_mean(stress_t_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_t_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f ", mean, std);
    mean = calc_mean(stress_r_array, MAX_NUM_EXPERIMENT);
    std = calc_std(stress_r_array, MAX_NUM_EXPERIMENT);
    fprintf(plot_fp2, "%0.2f %0.2f\n", mean, std);
  }

  fclose(plot_fp1);
  fclose(plot_fp2);
}

/************************************************/
/* process results from experiment 13	        */
/* varying percentage of link failures		*/
/************************************************/

void proc_13() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out13";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN];
  FILE *plot_fp1;
  float *perAD_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perAE_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perBD_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perBE_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perC_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, j, group_size = 256;

  int TAPESTRY_NODEID_SIZE = 6;         // 6
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  // int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;   // 4096
  
  for (i = 0; i < 4; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp13.%s.hop.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
  
    for (j = 0; j <= 50; j++) {
      exp13_14_read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, j,
		perAD_array, perAE_array, perBD_array, perBE_array,
		perC_array);
      mean = calc_mean(perAD_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perAD_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.2f %0.5f %0.5f ", j * 1.0 / 100, mean, std);
      mean = calc_mean(perAE_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perAE_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perBD_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perBD_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perBE_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perBE_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perC_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perC_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f\n", mean, std);
    }
    fclose(plot_fp1);
  }
}

/************************************************/
/* process results from experiment 14           */
/* varying percentage of node failures          */
/************************************************/

void proc_14() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char *topology_fn;
  char out_dir[MAX_FILENAME_LEN] = "out14";
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN];
  FILE *plot_fp1;
  float *perAD_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perAE_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perBD_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perBE_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *perC_array = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, j, group_size = 256;
  
  int TAPESTRY_NODEID_SIZE = 6;         // 6
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  // int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;   // 4096
  
  for (i = 0; i < 2; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp14.%s.hop.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
 
    for (j = 0; j <= 50; j++) {
      exp13_14_read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, j,
                perAD_array, perAE_array, perBD_array, perBE_array,
                perC_array);
      mean = calc_mean(perAD_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perAD_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%d %0.5f %0.5f ", j, mean, std);
      mean = calc_mean(perAE_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perAE_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perBD_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perBD_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perBE_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perBE_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f ", mean, std);
      mean = calc_mean(perC_array, MAX_NUM_EXPERIMENT);
      std = calc_std(perC_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%0.5f %0.5f\n", mean, std);
    }
    fclose(plot_fp1);
  }
}

/************************************************/
/* process results from experiment 15		*/
/* varying num_roots in tree partitioning       */
/************************************************/

void proc_15() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out15_tues";
  char *topology_fn;
  char plot_dir[MAX_FILENAME_LEN] = "plot";
  char plot_file1[MAX_FILENAME_LEN];
  char plot_file2[MAX_FILENAME_LEN];
  FILE *plot_fp1, *plot_fp2;
  float *ratio_array1 = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *ratio_array2 = (float *) malloc(sizeof(float) * MAX_NUM_EXPERIMENT);
  float *join_req_dev_array=(float*)malloc(sizeof(float)*MAX_NUM_EXPERIMENT);
  float mean, std;
  int i, num_roots;
  int group_size = 4096 - (1 + MAX_NUM_ROOTS);

  int TAPESTRY_NODEID_SIZE = 6;         // 6
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  // int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;   // 4096

  for (i = 0; i < 4; i++) {
    topology_fn = topology_fn_array[i];
    sprintf(plot_file1, "%s/exp15.%s.hop.out", plot_dir, topology_fn);
    plot_fp1 = fopen(plot_file1, "w");
    sprintf(plot_file2, "%s/exp15.%s.root.out", plot_dir, topology_fn);
    plot_fp2 = fopen(plot_file2, "w");

    for (num_roots = 0; num_roots < MAX_NUM_ROOTS; num_roots++) {
      exp15_read_delay(topology_fn, out_dir,
                TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, num_roots,
		ratio_array1, ratio_array2);
      mean = calc_mean(ratio_array1, MAX_NUM_EXPERIMENT);
      std = calc_std(ratio_array1, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, "%d %0.2f %0.2f", num_roots+1, mean, std);
      mean = calc_mean(ratio_array2, MAX_NUM_EXPERIMENT);
      std = calc_std(ratio_array2, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp1, " %0.2f %0.2f\n", mean, std);

      mean = group_size * 1.0 / (num_roots + 1);
      fprintf(plot_fp2, "%d %0.2f", num_roots+1, mean);
      exp15_read_root(topology_fn, out_dir,
		TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_SIZE, group_size, num_roots,
		join_req_dev_array); 
      mean = calc_mean(join_req_dev_array, MAX_NUM_EXPERIMENT);
      fprintf(plot_fp2, " %0.2f\n", mean); 
    }
    fclose(plot_fp1);
    fclose(plot_fp2);
  }
}

/************************************************/
/* MAIN						*/
/************************************************/

int main() {

  // experiment for 3 & 8
  // proc_3_8();

  // experiment for 5 & 10
  // proc_5_10();

  // experiment for 4 & 9
  // proc_4_9();

  // experiment for 11
  // proc_11();

  // experiment for 11
  // proc_11();

  // experiment for 12:  clustering receiver nodeIDs 
  // proc_12();

  // experiment for 13:  link failures 
  // proc_13();

  // experiment for 14:  node failures
  // proc_14();

  // experiment for 15:  tree partitioning 
  proc_15();

  return 0;
}
