#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <gb_graph.h>

#include "exp.h"
#include "stat.h"

/************************************************/
/* find out the worst case physical link        */
/* stress in tapestry multicast routing for     */
/* topology_fn in MAX_NUM_EXPERIMENT exps       */
/************************************************/

void read_stress(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size,
                float *stress_unicast_array, float *stress_tapestry_array,
		float *stress_ratio_array) {

  char param[MAX_FILENAME_LEN];
  char stress_unicast_fn[MAX_FILENAME_LEN];
  char stress_tapestry_fn[MAX_FILENAME_LEN];
  FILE *stress_unicast_fp, *stress_tapestry_fp;
  char line[MAX_LINE_LEN], str[MAX_LINE_LEN];
  int i_stress, max_stress, vnn;
      
  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
                        
    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size, vnn);
    sprintf(stress_unicast_fn,"%s.stress.unicast", param);
    sprintf(stress_tapestry_fn, "%s.stress.tapestry", param);

    // open, read, and close stress unicast file 
    stress_unicast_fp = fopen(stress_unicast_fn, "r");
    max_stress = INT_MIN;
    while (fgets(line, MAX_LINE_LEN, stress_unicast_fp)) {
      sscanf(line, "%s %d\n", str, &i_stress);
      if (i_stress > max_stress)
        max_stress = i_stress;
    }
    fclose(stress_unicast_fp);
    stress_unicast_array[vnn] = max_stress;

    // open, read, and close stress tapestry file 
    stress_tapestry_fp = fopen(stress_tapestry_fn, "r");
    max_stress = INT_MIN;
    while (fgets(line, MAX_LINE_LEN, stress_tapestry_fp)) {
      sscanf(line, "%s %d\n", str, &i_stress);
      if (i_stress > max_stress)
        max_stress = i_stress;
    }
    fclose(stress_tapestry_fp);
    stress_tapestry_array[vnn] = max_stress;

    stress_ratio_array[vnn] = stress_unicast_array[vnn] / 
				stress_tapestry_array[vnn];
  }
} 

/************************************************/
/* find out the 90 percentile RDP		*/
/* in tapestry multicast routing for		*/
/* topology_fn in MAX_NUM_EXPERIMENT exps       */
/************************************************/

void read_delay(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size,
                float *rdp_array) {

  char param[MAX_FILENAME_LEN];
  char hop_sp_fn[MAX_FILENAME_LEN];
  char hop_tapestry_fn[MAX_FILENAME_LEN];
  char hop_rdp_fn[MAX_FILENAME_LEN];
  FILE *hop_sp_fp, *hop_tapestry_fp, *hop_rdp_fp;
  int *hop_sp_array, *hop_tapestry_array;
  float *hop_rdp_array;
  int vnn;
  int n = group_size - 1;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {

    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size, vnn);
    sprintf(hop_sp_fn, "%s.hop.sp", param);
    sprintf(hop_tapestry_fn, "%s.hop.tapestry", param);
    sprintf(hop_rdp_fn, "%s.hop.rdp", param);

    // allocate memory for sp, tapestry, and rdp arrays
    hop_sp_array = malloc(sizeof(long) * n);
    hop_tapestry_array = malloc(sizeof(long) * n);
    hop_rdp_array = malloc(sizeof(float) * n);

    // open, read, and close sp, tapestry, and rdp files
    hop_sp_fp = fopen(hop_sp_fn, "r");
    hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
    hop_rdp_fp = fopen(hop_rdp_fn, "r");
    fread(hop_sp_array, sizeof(long), n, hop_sp_fp);
    fread(hop_tapestry_array, sizeof(float), n, hop_tapestry_fp);
    fread(hop_rdp_array, sizeof(float), n, hop_rdp_fp);
    fclose(hop_sp_fp);
    fclose(hop_tapestry_fp);
    fclose(hop_rdp_fp);

    rdp_array[vnn] = calc_percentile(hop_rdp_array, n, 0.9); 
  }
}

/************************************************/
/* find out the worst case physical link        */
/* stress in tapestry multicast routing for     */
/* topology_fn in MAX_NUM_EXPERIMENT exps       */
/************************************************/

void exp12_read_stress(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size, int num_clusters,
                float *stress_unicast_array, float *stress_tapestry_array,
                float *stress_ratio_array) {

  char param[MAX_FILENAME_LEN];
  char stress_unicast_fn[MAX_FILENAME_LEN];
  char stress_tapestry_fn[MAX_FILENAME_LEN];
  FILE *stress_unicast_fp, *stress_tapestry_fp;
  char line[MAX_LINE_LEN], str[MAX_LINE_LEN];
  int i_stress, max_stress, vnn;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {

    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE,
                        group_size, num_clusters, vnn);
    sprintf(stress_unicast_fn,"%s.stress.unicast", param);
    sprintf(stress_tapestry_fn, "%s.stress.tapestry", param);

    // open, read, and close stress unicast file
    stress_unicast_fp = fopen(stress_unicast_fn, "r");
    max_stress = INT_MIN;
    while (fgets(line, MAX_LINE_LEN, stress_unicast_fp)) {
      sscanf(line, "%s %d\n", str, &i_stress);
      if (i_stress > max_stress)
        max_stress = i_stress;
    }
    fclose(stress_unicast_fp);
    stress_unicast_array[vnn] = max_stress;

    // open, read, and close stress tapestry file
    stress_tapestry_fp = fopen(stress_tapestry_fn, "r");
    max_stress = INT_MIN;
    while (fgets(line, MAX_LINE_LEN, stress_tapestry_fp)) {
      sscanf(line, "%s %d\n", str, &i_stress);
      if (i_stress > max_stress)
        max_stress = i_stress;
    }
    fclose(stress_tapestry_fp);
    stress_tapestry_array[vnn] = max_stress;

    stress_ratio_array[vnn] = stress_unicast_array[vnn] /
                                stress_tapestry_array[vnn];
  }
}

/************************************************/
/* find out the 90 percentile RDP               */
/* in tapestry multicast routing for            */
/* topology_fn in MAX_NUM_EXPERIMENT exps       */
/************************************************/

void exp12_read_delay(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size, int num_clusters,
                float *rdp_array) {

  char param[MAX_FILENAME_LEN];
  char hop_sp_fn[MAX_FILENAME_LEN];
  char hop_tapestry_fn[MAX_FILENAME_LEN];
  char hop_rdp_fn[MAX_FILENAME_LEN];
  FILE *hop_sp_fp, *hop_tapestry_fp, *hop_rdp_fp;
  int *hop_sp_array, *hop_tapestry_array;
  float *hop_rdp_array;
  int vnn;
  int n = group_size - 1;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {

    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE,
                        group_size, num_clusters, vnn);
    sprintf(hop_sp_fn, "%s.hop.sp", param);
    sprintf(hop_tapestry_fn, "%s.hop.tapestry", param);
    sprintf(hop_rdp_fn, "%s.hop.rdp", param);

    // allocate memory for sp, tapestry, and rdp arrays
    hop_sp_array = malloc(sizeof(long) * n);
    hop_tapestry_array = malloc(sizeof(long) * n);
    hop_rdp_array = malloc(sizeof(float) * n);

    // open, read, and close sp, tapestry, and rdp files
    hop_sp_fp = fopen(hop_sp_fn, "r");
    hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
    hop_rdp_fp = fopen(hop_rdp_fn, "r");
    fread(hop_sp_array, sizeof(long), n, hop_sp_fp);
    fread(hop_tapestry_array, sizeof(float), n, hop_tapestry_fp);
    fread(hop_rdp_array, sizeof(float), n, hop_rdp_fp);
    fclose(hop_sp_fp);
    fclose(hop_tapestry_fp);
    fclose(hop_rdp_fp);

    rdp_array[vnn] = calc_percentile(hop_rdp_array, n, 0.9);
  }
}

/************************************************/
/* find out the AD, AE, BD, BE, C percentage	*/
/* distribution in tapestry multicast routing	*/
/* for topology_fn in MAX_NUM_EXPERIMENT exps	*/ 
/************************************************/
  
void exp13_14_read_delay(char *topology_fn, char *out_dir,    
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size, int percentage,
                float *perAD_array, float *perAE_array,
		float *perBD_array, float *perBE_array,
                float *perC_array) {
  char param[MAX_FILENAME_LEN]; 
  char hop_sp_fn[MAX_FILENAME_LEN];
  char hop_tapestry_fn[MAX_FILENAME_LEN]; 
  FILE *hop_sp_fp, *hop_tapestry_fp;
  int *hop_sp_array, *hop_tapestry_array;
  int vnn, i;              
  int n = group_size - 1;
  int numAD, numAE, numBD, numBE, numC;
    
  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    
    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE,
                        group_size, percentage, vnn);
    sprintf(hop_sp_fn, "%s.hop.sp", param);
    sprintf(hop_tapestry_fn, "%s.hop.tapestry", param);
    
    // allocate memory for sp, tapestry, and rdp arrays
    hop_sp_array = malloc(sizeof(long) * n);
    hop_tapestry_array = malloc(sizeof(long) * n);
    
    // open, read, and close sp, tapestry, and rdp files
    hop_sp_fp = fopen(hop_sp_fn, "r");
    hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
    fread(hop_sp_array, sizeof(long), n, hop_sp_fp);
    fread(hop_tapestry_array, sizeof(float), n, hop_tapestry_fp);
    fclose(hop_sp_fp);
    fclose(hop_tapestry_fp);
   
    numAD = numAE = numBD = numBE = numC = 0;
    for (i = 0; i < n; i++) {
      if (hop_sp_array[i] > 0) {
        if (hop_tapestry_array[i] > 0)
          numAD++;
        else if (hop_tapestry_array[i] == -1)
          numAE++; 
      }
      else if (hop_sp_array[i] == 0) {
        if (hop_tapestry_array[i] > 0)
          numBD++;
        else if (hop_tapestry_array[i] == -1)
          numBE++;
      }
      else if (hop_sp_array[i] == -1) {
        numC++;
      }
    }
    perAD_array[vnn] = numAD * 1.0 / n;
    perAE_array[vnn] = numAE * 1.0 / n;
    perBD_array[vnn] = numBD * 1.0 / n;
    perBE_array[vnn] = numBE * 1.0 / n;
    perC_array[vnn] = numC * 1.0 / n;
  } 
}

/************************************************/
/* find out the 90 percentile RDP               */
/* in tapestry multicast routing for            */
/* topology_fn in MAX_NUM_EXPERIMENT exps       */
/************************************************/

void exp15_read_delay(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE, int group_size, int num_roots,
                float *ratio_array1, float *ratio_array2) {

  char param[MAX_FILENAME_LEN];
  char hop_sp_fn[MAX_FILENAME_LEN];
  char hop_tapestry_fn[MAX_FILENAME_LEN];
  FILE *hop_sp_fp, *hop_tapestry_fp;
  int *hop_sp_array1, *hop_sp_array2;
  int *hop_tapestry_array1, *hop_tapestry_array2;
  float *sp_ratio_array, *tapestry_ratio_array;
  int vnn;
  int n = group_size, i;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {

    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size, 1, vnn);
    sprintf(hop_sp_fn, "%s.hop.sp", param);
    hop_sp_array1 = malloc(sizeof(long) * n);
    hop_sp_fp = fopen(hop_sp_fn, "r");
    fread(hop_sp_array1, sizeof(float), n, hop_sp_fp);
    fclose(hop_sp_fp);

    sprintf(hop_tapestry_fn, "%s.hop.tapestry", param);
    hop_tapestry_array1 = malloc(sizeof(long) * n);
    hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
    fread(hop_tapestry_array1, sizeof(float), n, hop_tapestry_fp);
    fclose(hop_tapestry_fp);

    // initialize filenames
    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size,num_roots+1,vnn);
    sprintf(hop_sp_fn, "%s.hop.sp", param);
    hop_sp_fp = fopen(hop_sp_fn, "r");
    hop_sp_array2 = malloc(sizeof(long) * n);
    fread(hop_sp_array2, sizeof(float), n, hop_sp_fp);
    fclose(hop_sp_fp);

    sprintf(hop_tapestry_fn, "%s.hop.tapestry", param);
    hop_tapestry_fp = fopen(hop_tapestry_fn, "r");
    hop_tapestry_array2 = malloc(sizeof(long) * n);
    fread(hop_tapestry_array2, sizeof(float), n, hop_tapestry_fp);
    fclose(hop_tapestry_fp);

    sp_ratio_array = (float *) malloc(sizeof(float) * n);
    tapestry_ratio_array = (float *) malloc(sizeof(float) * n);
    for (i = 0; i < n; i++) {
      sp_ratio_array[i] = (float) (hop_sp_array2[i] * 1.0 /
					hop_sp_array1[i]);
      tapestry_ratio_array[i] = (float) (hop_tapestry_array2[i] * 1.0 /
					hop_tapestry_array1[i]);
    }
    ratio_array1[vnn] = calc_mean(sp_ratio_array, n); 
    ratio_array2[vnn] = calc_mean(tapestry_ratio_array, n); 
  }
}

/************************************************/
/************************************************/

void exp15_read_root(char *topology_fn, char *out_dir,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_NAMESPACE_SIZE,
		int group_size, int num_roots, float *join_req_dev_array) {

  char param[MAX_FILENAME_LEN];
  char root_fn[MAX_FILENAME_LEN];
  FILE *root_fp;
  int i, vnn;
  long *join_req_array;
  float *join_req_float_array;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {

    sprintf(param, "%s/%s.%dx%dx%dx%dx%d.%d", out_dir, topology_fn,
                        TAPESTRY_DIGIT_SIZE, TAPESTRY_NODEID_SIZE,
                        TAPESTRY_NAMESPACE_SIZE, group_size, 1, vnn);
    sprintf(root_fn, "%s.root", param);
    join_req_array = (long *) malloc(sizeof(long) * (num_roots + 1));
    join_req_float_array = (float *) malloc(sizeof(float) * (num_roots + 1));
    root_fp = fopen(root_fn, "r");
    fread(join_req_array, sizeof(long), num_roots+1, root_fp);
    fclose(root_fp);

    for (i = 0; i < num_roots; i++) {
      join_req_float_array[i] = (float) join_req_array[i];
    }
    join_req_dev_array[vnn] = calc_std(join_req_float_array, num_roots + 1);
  }
}
