#include <math.h>
#include "util.h"

/************************************************/
/* calculate mean of a float array of size n	*/
/************************************************/

float calc_mean(float *array, int n) {
  float sum;
  int i;

  sum = 0.0;
  for (i = 0; i < n; i++)
    sum += array[i];

  return (sum / n);
}

/************************************************/
/* calculate standard deviation of a 		*/
/* float array of size n			*/
/************************************************/

float calc_std(float *array, int n) {
  float sum1, sum2, mean, S2, S;
  int i;

  sum1 = 0.0;
  for (i = 0; i < n; i++)
    sum1 += array[i];

  mean = sum1 / n;

  sum2 = 0.0;
  for (i = 0; i < n; i++)
    sum2 += (array[i] - mean) * (array[i] - mean);
  S2 = sum2 / n;                  // biased estimator for sigma^2
  S = sqrt(S2);

  return S;
}
/************************************************/
/* calculate confidence interval for mean,	*/
/* standard deviation estimated			*/
/* assumptions:					*/
/* 1. normal distribution, else sample size 	*/
/* should be at least 25			*/
/************************************************/

float calc_ci(int *array, int n) {
  float mean, sum1, sum2, S2, S, sample_standard_error, confidence_interval;
  int i;

  // degrees of freedom = sample size (N) - 1 = 25 - 1 = 24
  // 95% confidence interval
  float t = 2.064;

  sum1 = 0.0;
  for (i = 0; i < n; i++)
    sum1 += array[i];

  mean = sum1 / n;

  sum2 = 0.0;
  for (i = 0; i < n; i++)
    sum2 += (array[i] - mean) * (array[i] - mean);
  S2 = sum2 / (n - 1);                  // unbiased estimator for sigma^2
  S = sqrt(S2);
  sample_standard_error = S / sqrt(n);
  confidence_interval = mean + t * sample_standard_error;

  return confidence_interval;
}

/************************************************/
/* calculate the midpoint between (x1, y1) and	*/
/* (x2, y2)					*/
/************************************************/

void midpoint(float x1, float y1, float x2, float y2, float *mx, float *my) {
  *mx = (x1 + x2) / 2;
  *my = (y1 + y2) / 2;
}

/************************************************/
/* calculate the (percent)th percentile of a	*/
/* float array of size n			*/
/************************************************/

float calc_percentile(float *array, int n, float percent) {
  float percentile[n];
  int i;

  bubblesort_f(array, n);

  for (i = 0; i < n; i++) {
    percentile[i] = (i+1) / n; 
  }
  
  for (i = 0; i < n; i++) {
    if (percentile[i] > percent)
      return ((array[i-1] + array[i]) / 2); 
  }      
  return 0.0;
}
