#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define MAX_LINE_LEN 256
#define MAX_FILENAME_LEN 128

int main(int argc, char **argv) {
  char topology_fn[MAX_FILENAME_LEN];
  char stress_unicast_fn[MAX_FILENAME_LEN];
  char stress_tapestry_fn[MAX_FILENAME_LEN];
  FILE *stress_unicast_fp, *stress_tapestry_fp;
  char line[MAX_LINE_LEN], str[MAX_LINE_LEN];
  int stress, min_stress, max_stress, sum_stress, i, count;
  float avg_stress;

  if (argc < 2)
    exit(EXIT_FAILURE);

  // initialize filenames
  strcpy(topology_fn, argv[1]);
  sprintf(stress_unicast_fn,"%s.stress.unicast",topology_fn);
  sprintf(stress_tapestry_fn,"%s.stress.tapestry",topology_fn);

  // open, read, and close stress unicast file
  stress_unicast_fp = fopen(stress_unicast_fn, "r");
  min_stress = INT_MAX;
  max_stress = INT_MIN;
  sum_stress = 0;
  count = 0;
  while (fgets(line, MAX_LINE_LEN, stress_unicast_fp)) { 
    sscanf(line, "%s %d\n", str, &stress);
    if (stress < min_stress)
      min_stress = stress;
    if (stress > max_stress)
      max_stress = stress;
    sum_stress += stress;
    count++;
  }
  avg_stress = sum_stress * 1.0 / count;
  printf("unicast best case physical link stress = %d\n", min_stress);
  printf("unicast average case physical link stress = %.2f\n", avg_stress);
  printf("unicast worst case physical link stress = %d\n", max_stress);
  for (i = min_stress; i <= max_stress; i++) {
    rewind(stress_unicast_fp);
    count = 0;
    while (fgets(line, MAX_LINE_LEN, stress_unicast_fp)) {
      sscanf(line, "%s %d\n", str, &stress);
      if (stress == i)
        count++;
    }
    if (count > 0)
      printf("%d %d\n", i, count);
  }
  fclose(stress_unicast_fp);;

  // open, read, and close stress tapestry file
  stress_tapestry_fp = fopen(stress_tapestry_fn, "r");
  min_stress = INT_MAX;
  max_stress = INT_MIN;
  sum_stress = 0;
  count = 0;
  while (fgets(line, MAX_LINE_LEN, stress_tapestry_fp)) { 
    sscanf(line, "%s %d\n", str, &stress);
    if (stress < min_stress)
      min_stress = stress;
    if (stress > max_stress)
      max_stress = stress;
    sum_stress += stress;
    count++;
  }
  avg_stress = sum_stress * 1.0 / count;
  printf("tapestry best case physical link stress = %d\n", min_stress);
  printf("tapestry average case physical link stress = %.2f\n", avg_stress);
  printf("tapestry worst case physical link stress = %d\n", max_stress);
  for (i = min_stress; i <= max_stress; i++) {
    rewind(stress_tapestry_fp);
    count = 0;
    while (fgets(line, MAX_LINE_LEN, stress_tapestry_fp)) {
      sscanf(line, "%s %d\n", str, &stress);
      if (stress == i)
        count++;
    }
    if (count > 0)
      printf("%d %d\n", i, count);
  }
  fclose(stress_tapestry_fp);

  return 0;
}
