#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_dijk.h>
#include <gb_save.h>
#include <gb_basic.h>
#include "graph.h"

#define MAX_FILENAME_LEN 128

void print_all_arcs(Graph *g) {
  Vertex *v;
  Arc *a;
  int count = 0; 

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    printf("%s:", v->name);
    for (a = v->arcs; a; a = a->next) {
      printf(" %s(%ld)", a->tip->name, a->len);
    }
    printf("\n");
  }
}

void print_stub_size(Graph *g) {
  Vertex *v;
  int len;
  char *ptr = NULL;
  char vname[256], stub_size[256];

  v = g->vertices;
  while (v < g->vertices + g->n) {
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      // printf("%s ", v->name);
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) {
      ptr = strchr(vname, '.');
      printf("%s - %s (%d)\n", (v-atoi(ptr+1)-1)->name, 
				(v-1)->name,atoi(ptr+1)+1);
    }
    strcpy(vname, v->name);
    ptr = strrchr(vname, '.');
    *ptr = '\0';
    len = strlen(vname);
    while ((v+1) < g->vertices+g->n && strncmp(vname,(v+1)->name,len)==0) {
      v++;
    }
    ptr = strrchr(v->name, '.');
    strcpy(stub_size, ptr+1);
    printf("%s - %s (%d)\n", (v-atoi(stub_size))->name, v->name,
						atoi(stub_size)+1);
    v++;
  } 
}

void swap(int *a, int *b) {
  int tmp = *a;
  *a = *b;
  *b = tmp;
}

void bubblesort (int *array, int size) {
  int i, j;

  for (i = 0; i < size; i++)
    for (j = size - 1; j > i; j--)
      if (array[j] < array[j-1])
        swap(&array[j], &array[j-1]);
}

void pick_clusters(Graph *g) {
  int num_stubs = 1520;
  int stub_array[1520];
  int num_clusters = 256;
  int cluster_array[256];
  int i, j, index, len;
  Vertex *v;
  char vname[256], stub_size[256], *ptr;

  for (i = 0; i < num_stubs; i++) {
    stub_array[i] = 0;
  }
  for (i = 0; i < num_clusters; i++) {
    index = gb_unif_rand(1520);
    while (stub_array[index] == 1) {
      index = gb_unif_rand(1520);
    }
    stub_array[index] = 1;
    cluster_array[i] = index;
    printf("%d\n", cluster_array[i]);
  }
  printf("*****\n");

  bubblesort(cluster_array, 256);
  for (i = 0; i < num_clusters; i++) {
    printf("%d %d\n", i, cluster_array[i]);
  }

  i = 0;
  j = 0;
  v = g->vertices;
  printf("after \n");
  while (i < num_clusters && v < g->vertices + g->n) {    
    printf("entering i = %d\n", i);
    while (v < g->vertices + g->n && strchr(v->name, 'T')) {
      v++;
    }
    strcpy(vname, (v-1)->name);
    if (strchr(vname, 'T')) {
      ptr = strchr(vname, '.');  
      printf("%d %s - %s (%d)\n", j, (v-atoi(ptr+1)-1)->name,
                                (v-1)->name,atoi(ptr+1)+1);
      if (j == cluster_array[i]) {
        printf("i = %d, domain %d\n", i, j);
        i++;
      }
      j++;
    }
    if (i < num_clusters) {
      strcpy(vname, v->name);     
      ptr = strrchr(vname, '.');
      *ptr = '\0';
      len = strlen(vname);
      while ((v+1) < g->vertices+g->n && strncmp(vname,(v+1)->name,len)==0) {
        v++;
      }
      ptr = strrchr(v->name, '.');
      strcpy(stub_size, ptr+1);
      printf("%s - %s (%d)\n", (v-atoi(stub_size))->name, v->name,
                                                atoi(stub_size)+1);
      if (j == cluster_array[i]) {
        printf("i = %d, domain %d\n", i, j);
        i++;
      }
      j++;
      v++;
    }
  }
}

Graph *remove_vertices(Graph *g) {
  Graph *new_graph;
  Vertex *v;

  v = g->vertices;
  v->ind = 1;
  (v+4)->ind = 1;
  (v+6)->ind = 1;
  (v+9)->ind = 1;
  new_graph = induced(g, NULL, 0, 0, 0); 
  return new_graph;
}

void inject_link_faults(Graph *g, int num_broken_links) {
  int i;
  Vertex *v, *u;
  Arc *first_arc, *a, *pred_a;
  
  for (i = 0; i < num_broken_links; i++) {
    first_arc = NULL;
    while (first_arc == NULL) {
      v = g->vertices + gb_unif_rand(g->n);
      first_arc = v->arcs;
    }
    v->arcs = first_arc->next;
    u = first_arc->tip;
    pred_a = NULL;
    for (a = u->arcs; a; a = a->next) {
      if (a->tip == v) {
        if (a == u->arcs) {
          u->arcs = a->next;
        } else {
          pred_a->next = a->next;
        }
        break;
      }
      pred_a = a;
    }
  }
}

int main(int argc, char **argv) {
  Graph *g, *new_graph;
  char topology_fn[MAX_FILENAME_LEN];
  long distance;

  gb_init_rand(0);     // initialize the random number generator

  if (argc < 2)
    exit(EXIT_FAILURE);

  // initialize filenames
  sprintf(topology_fn, "%s", argv[1]);

  g = restore_graph(topology_fn);
  print_all_arcs(g);

/*
  new_graph = remove_vertices(g);
  printf("\n");
  printf("%ld\n", new_graph->vertices->picked_t);
  print_all_arcs(new_graph);
*/

  distance = dijkstra(g->vertices, g->vertices+1, g, NULL);
  printf("%ld\n", (g->vertices+1)->dist);
  printf("%ld\n", distance);

  inject_link_faults(g, 5);
  printf("\n");
  print_all_arcs(g);

  distance = dijkstra(g->vertices, g->vertices+1, g, NULL);
  printf("%ld\n", (g->vertices+1)->dist);
  printf("%ld\n", distance);

  // print_stub_size(g);

  // pick_clusters(g);

  return 0;
}
