#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "exp.h"
#include "pick.h"
#include "graph.h"
#include "hop.h"
#include "util.h"

/************************************************/
/* experiment 3 & 8				*/
/* varying multicast group size			*/
/************************************************/

void exp_3_8() {
  char *topology_fn_array[9] = {"ts5000.gb",
	"ts10000.gb", "ts20000.gb", "ts30000.gb", "ts40000.gb", "ts50000.gb", 
	"ti5000.gb", "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out3_8";
  int i, group_size;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 6;		// 6
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;	// 4096

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 1; i < 5; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      for (group_size = 16; group_size <= 4096; group_size *= 2) {
        exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
		group_size, vnn);
      }
    }
  }
}

/************************************************/
/* experiment 5 & 10                            */
/* varying tapestry namespace size		*/
/************************************************/

void exp_5_10() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out5_10";
  int i, group_size = 64;
  int vnn;
  
  int TAPESTRY_NODEID_SIZE;		// 3, 4, 5, 6 
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE;		// 64, 256, 1024, 4096 

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 4; i++) {
      for (TAPESTRY_NODEID_SIZE = 3; TAPESTRY_NODEID_SIZE <= 6; 
						TAPESTRY_NODEID_SIZE++) {
        TAPESTRY_NAMESPACE_SIZE = pow(TAPESTRY_NAMESPACE_BASE, 
					TAPESTRY_NODEID_SIZE);
        g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
	exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
	
      }
    }
  }
}

/************************************************/
/* experiment 4 & 9				*/
/* varying topology size			*/
/************************************************/

void exp_4_9() {
  char *topology_fn_array[14] = {"ts100.gb", "ts500.gb", "ts1000.gb",
		"ts2000.gb", "ts3000.gb", "ts4000.gb", "ts5000.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out4_9";
  int i, group_size = 64;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 3;
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 64;

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 7; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
    }
  }
}

/************************************************/
/* experiment 11				*/
/* varying tapestry namespace base (fanout)	*/
/************************************************/

void exp_11() {
  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  int group_size = 64;
  int vnn;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE;
  Graph *g;

//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  for (vnn = 5; vnn < 10; vnn++) {
    TAPESTRY_NODEID_SIZE = 3; 
    TAPESTRY_DIGIT_SIZE = 4;
    TAPESTRY_NAMESPACE_BASE = 16;	// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 4; 
    TAPESTRY_DIGIT_SIZE = 3;
    TAPESTRY_NAMESPACE_BASE = 8;		// octal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 6; 
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;		// quad digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 12; 
    TAPESTRY_DIGIT_SIZE = 1;
    TAPESTRY_NAMESPACE_BASE = 2;		// binary digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
  }
}

void exp_archival(int switchy, int hiloload) {
  char topology_fn[MAX_FILENAME_LEN] = "ts5000.gb";
  int mask, mask2, r_id, digit, i, j, k, RUNS=100;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int OBJECTS = 32;
  int THRESHOLD = OBJECTS / 2;
  int MAX_EXPERIMENTS = 1;
  int NUMBER_LOADED_SERVERS = 16;
  int LOAD_LATENCY_LOW = 1;
  int LOAD_LATENCY_HIGH = 6;
  float FAILRATE;

  Vertex *client, *root, *v1, *v2, *r, **routing_table;
  Vertex ** servers;
  int serverload[OBJECTS];

  if (hiloload == 1) {
    LOAD_LATENCY_LOW = 5;
    LOAD_LATENCY_HIGH = 10;
  }

  servers = (Vertex **) malloc(sizeof(Vertex *) *OBJECTS);
  
  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;

  for (k = 0; k < MAX_EXPERIMENTS; k++) {
    g = exp_init (topology_fn, TAPESTRY_NODEID_SIZE,
		  TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    clear_picked_m_field(g);
    root = pick_0(g);
    
    r = root;
    for (j=0; j<OBJECTS; j++) {
      servers[j] = pick_0(g);
      mask2 = 1 << j;
      root->picked_m = root->picked_m | mask2;
      servers[j]->picked_m = servers[j]->picked_m | mask2;
      if (j<NUMBER_LOADED_SERVERS) 
	serverload[j] = LOAD_LATENCY_LOW;
      else serverload[j] = LOAD_LATENCY_HIGH;
      
      v1 = servers[j];
      for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
	tapestry_nodeID_atoi(r->nodeid, &r_id, 
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE *i +digit];
	v2->picked_m = v2->picked_m |mask2;
	v1= v2;
      }
    }
    
    client = pick_0(g);
    if (switchy == 13) {
      printf("16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\n");
      for (j = 0; j<RUNS; j++) {
	for (i=THRESHOLD; i<OBJECTS; i++) {  //VARYING # OF REQUESTS
	  hop_calc_archival(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			    TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			    OBJECTS, THRESHOLD, i, serverload, 0.0);
	}
	printf("\n");
      }
    } else if (switchy == 14) {
      for (FAILRATE=0.0; FAILRATE < 0.121; FAILRATE=FAILRATE+0.02) {
	printf("--------------FailRate:%f---------------------\n",
	       FAILRATE);
	for (j = 0; j<RUNS; j++) {
	  for (i=THRESHOLD; i<OBJECTS; i++) {  //VARYING # OF REQUESTS
	    hop_calc_archival(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			      TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			      OBJECTS, THRESHOLD, i, serverload,
			      FAILRATE);
	  }
	  printf("\n");
	}
      }
    } // end else if
  }
}


void exp_12() {
  char topology_fn[MAX_FILENAME_LEN] = "mb022399.gb";
  //  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  //  int group_size = 64;
  int vnn, mask, mask2, r_id, digit, i, j, RUNS=200;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int OBJECTS = 5;
  Vertex *client, *root, *v1, *v2, *r, **routing_table;
  Vertex ** servers;
 
  servers = (Vertex **) malloc(sizeof(Vertex *) * OBJECTS);
//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  for (vnn = 0; vnn < 5; vnn++) {
    TAPESTRY_NODEID_SIZE = 4;
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 256;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(g);
    root = pick_0(g);
    root->picked_m = 1;
    r = root;

    for (j=0; j<OBJECTS; j++) {
      servers[j] = pick_0(g);
      mask2 = 1 << j;
      root->picked_m = root->picked_m | mask2;
      servers[j]->picked_m = servers[j]->picked_m | mask2;
      // mark the nodes from server up to object root
      v1 = servers[j];
      for (i=0; i< TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	
	mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i +digit];
	v2->picked_m = v2->picked_m | mask2;
	v1 = v2;
      }
    }

    printf("Run #%d: \n", vnn);
    for (i=0; i< RUNS; i++) {
      client = pick_0(g);
      printf("%f\n", 
	     hop_calc_loc_objs(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			       TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			       OBJECTS));
/* 
       hop_calc_loc_IP(g, client, server);
       hop_calc_loc_TAP(g, client, root, server, TAPESTRY_NODEID_SIZE,
 			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
       hop_calc_loc_PAT(g, client, root, server, TAPESTRY_NODEID_SIZE,
 			   TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
       printf("%d,     %d,     %d\n", a, b, c);
*/
    }
  }
}


/************************************************/
/* MAIN						*/
/************************************************/

int main() {

  //  int i;
  //  float random;

  gb_init_rand(time(NULL));     // initialize the random number generator

  // experiment for 3 & 8
  // exp_3_8();

  // experiment for 5 & 10
  // exp_5_10();

  // experiment for 4 & 9
  // exp_4_9();

  // experiment for 11
  exp_archival(14, 0);       //exp_archival(exp#, load: 1=high, 0=low)

  /*  for (i=0; i<10000; i++) {
    random = ((float) rand()) / RAND_MAX;
    printf("%f\n", -1* log(random));
  }
  */
  return 0;
}
