#include <gb_graph.h>
#include <gb_dijk.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>

#include "graph.h"
#include "hop.h"
#include "nodeid.h"
#include "util.h"

/**************
 ** calculate IP distance in locating objects
 **/
int hop_calc_loc_IP(Graph *g, Vertex *client, Vertex *server) {
  dijkstra(client, server, g, NULL);
  return server->dist;
}

/**************
 ** calculate IP distance in locating objects
 **/
int hop_calc_loc_IP_print(Graph *g, Vertex *client, Vertex *server) {
  dijkstra(client, server, g, NULL);
  printf("%ld     ", server->dist);
  return server->dist;
}

/*********
 ** return whether the number x has a 1 in the lower n bits
 **/
int lowerNBits(unsigned x, int n) {
  int i;
  //  printf ("Calling lowerNBits: %lu, %d\n", x, n);
  for (i = 0; i<=n && x!=0 ; i++, x = x>>1) {
    if (x & 01) return i;
  }
  return -1;
}

/**************
 ** calculate avg time to get threshold fragments in the archival system
 *************/
void hop_calc_archival(Graph *g, Vertex *client, Vertex *root, 
		       Vertex **servers, int TAPESTRY_NODEID_SIZE,
		       int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE,
		       int OBJECTS, int THRESHOLD, int REQUESTS,
		       int serverload[], float frate) {
  int i, j, k, l, distance, mask, r_id, digit, d, bits;
  int mask2, r_id2, digit2, mark, totald, minindex;
  float minval;
  float distances[OBJECTS];
  Vertex *v1, *v2, *r, *v12, *v22, *r2;
  Vertex ** routing_table, **routing_table2;
  float sorted[THRESHOLD];
  float random;
  int newbits;
  int locald[OBJECTS];
  int closestd, closestindex;

  v1 = client;
  r = root;
  distance = 0;
  mark = 0;  //# of servers we've contacted
  totald = 0; //total net distance traveled by all objects requested

  for (i=0; i<OBJECTS; i++) {
    distances[i] = -1;
    //initialize distances to -1 to mark they haven't been touched
  }
  
  for (j =0; j<TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE*j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE *j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
    if (v1->picked_m != 0) {
      bits = v1->picked_m;
      
      for (i=0; i < OBJECTS; i++) {
	locald[i] = -1;
      }
      //      printf("new indices at: ");
      for (i=0, newbits=0; bits !=0 && i<OBJECTS; i++, bits = bits >> 1) {
	if ((bits & (~(~0 << 1))) == 0) continue;
	// if distance already there, we've done it before, so skip
	if (distances[i] != -1) continue;
	locald[i] = hop_calc_loc_IP(g, v1, servers[i]);
	//printf("%d-%f ",i,distances[i]);
	newbits++;
      }
      //printf("\n");

      for (i=0; i<newbits; i++) {
	closestindex = -1;
	closestd = INT_MAX;
	for (l = 0; l< OBJECTS; l++) {
	  if (locald[l] == -1 ) continue;
	  if (locald[l] < closestd) {
	    closestd = locald[l];
	    closestindex = l;
	  }
	}
	//printf("got closest obj at locald index %d\n", closestindex);
	//	printf("%d ",locald[closestindex]);
	//if (closestindex == -1) break; // no more new servers in this node
	locald[closestindex] = -1;  //wipe it out after we found it 

	d = distance;

	v12 = v1;
	r2 = servers[closestindex];
	for (k =0; k<TAPESTRY_NODEID_SIZE; k++) {
	  routing_table2 = (Vertex **) v12->rtable;
	  mask2 = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*k);
	  tapestry_nodeID_atoi(r2->nodeid, &r_id2,
			       TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	  digit2 = (r_id2 & mask2) >> (TAPESTRY_DIGIT_SIZE *k);
	  v22 = routing_table2[TAPESTRY_NAMESPACE_BASE*k + digit2];
	  dijkstra(v12, v22, g, NULL);
	  d += v22->dist;
	  v12 = v22;
	}
	//now need to add standard deviations and server load
	totald+=d;
	if ((frate * d) >= ((float) rand()) / RAND_MAX) {
	  // if link failure, then repeat last hop to server
	  d = 0 - d;
	}
	random = ((float) rand()) / RAND_MAX;
	//printf("distance[%d] = %f\n", closestindex, distances[closestindex]);
	if (d>0) 
	  distances[closestindex] = 
	    ((float) (d * -1 * log(random)))  + serverload[i];
	else distances[closestindex] = -1;
	//printf("distance[%d] = %f\n", closestindex, distances[closestindex]);
	mark += 1;
	if (mark >= REQUESTS) break;
      } //end while loop

    } // end if
    if (mark >= REQUESTS) break;
  } //end for (j=0;)

  //  printf("\n");
  //now done with # of requests, analyze and print results
  //for graph 1, we want the maximum value from the smallest THRESHOLD values
  //for graph 2, we want the total distance of all paths taken
  for (i=0; i<THRESHOLD-1; i++) {
    //find smallest distance/time in list of responses
    minval = 0.0+INT_MAX;
    minindex = -1;
    for (j=0; j<OBJECTS; j++) {
      if (distances[j] > 0 && distances[j] < minval) {
	minval = distances[j];
	minindex = j;
      }
    }
    if (minindex == -1) {
      // ran out of valid distances before reaching threshold, failed
      printf("-1.0,%d ",totald);
      return;
    }
    distances[minindex] = -1;
    sorted[i] = minval;
  }   //ok, we've now found threshold -1 smallest items, next one will be it
  minval = (float) INT_MAX;
  minindex = -1;
  for (j=0; j<OBJECTS; j++) {
    if (distances[j] > 0 && distances[j] < minval) {
      minval = distances[j];
      minindex = j;
    }
  }
  if (minindex == -1) printf("-1.0,%d ",totald);
  else printf("%f,%d ", minval, totald);
  //  printf ("%d requests: ", REQUESTS);
/*  for (i=0; i< THRESHOLD-1; i++) {
    printf("%f ", sorted[i]);
  }
*/
}

void hop_calc_loc_planes(Graph *g, Vertex *client, Vertex **roots, 
			 Vertex *server, int TAPESTRY_NODEID_SIZE,
			 int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE,
			 int PLANES) {
  int i, j, distance, mask, r_id, digit;
  int mask2, totald, minindex;
  float minval;
  float distances[PLANES];
  float d[PLANES];
  Vertex *v1, *v2, *r;
  Vertex ** routing_table;
  float random;
  
  for (i=0; i<PLANES; i++) {
    distances[i] = -1;
    //initialize distances to -1 to mark they haven't been touched
  }
  for (i = 0; i < PLANES; i++) {
    v1 = client;
    r = roots[i];
    distance = 0;
    mask2 = 1 << i;
  
    for (j =0; j<TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE*j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
			   TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE *j +digit];
      dijkstra(v1, v2, g, NULL);
      distance += v2->dist;
      v1 = v2;
      if ((v1->picked_m & mask2) != 0) {
	break;
      }
    }
    
    r = server;
    for (j=0; j<TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE*j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
			   TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE *j +digit];
      dijkstra(v1, v2, g, NULL);
      distance += v2->dist;
      v1 = v2;
    }

    d[i] = distance;
    random = ((float) rand()) / RAND_MAX;
    //printf("distance[%d] = %f\n", closestindex, distances[closestindex]);
    distances[i] = (float) (distance * -1 * log(random));
    //printf("distance[%d] = %f\n", closestindex, distances[closestindex]);
  } //end for (i=0;)

  //now done with # of requests, analyze and print results
  //for graph 1, we want the maximum value from the smallest THRESHOLD values
  //for graph 2, we want the total distance of all paths taken
  for (j=0; j < PLANES; j++) {
    minval = INT_MAX;
    minindex = -1;
    totald = 0; //total net distance traveled by all objects requested
    for (i=0; i<=j; i++) {
      //find smallest distance/time in list of responses
      totald += d[i];
      if (distances[i] < minval) {
	minval = distances[i];
	minindex = i;
      }
    }
    printf("   %f, %d ", minval, totald);
    //  printf ("%d requests: ", REQUESTS);
  }

}

/**************
 ** calculate tapestry distance in locating closest object of many
 **/
float hop_calc_loc_objs(Graph *g, Vertex *client, Vertex *root, 
		       Vertex **servers,
		       int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
		       int TAPESTRY_NAMESPACE_BASE, int OBJECTS) {
  int i, j, distance, mask, r_id, digit, d1, d2;
  Vertex *v1, *v2, *r;
  Vertex **routing_table;
  int distances[OBJECTS];
  int nearest, t;
  Vertex *nearestS;

  for (i = 0; i< OBJECTS; i++) {
    distances[i] = hop_calc_loc_IP(g, client, servers[i]);
  }

  v1 = client;
  r = root;
  distance = 0;
  
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
    if (v1->picked_m != 0) {
      nearest = 50000;
      nearestS = NULL;
      for (i =0; i<OBJECTS; i++) {
	if ((t=hop_calc_loc_IP(g, v1, servers[i])) <nearest) {
	  nearestS = servers[i];
	  d1 = i;
	}
      }
      break;
    }
  }
  
  //  dijkstra(v1, server, g, NULL);
  //  d = distance + server->dist;
  //  if (d == -1) {
  //    printf("\nSHOULD NEVER HAPPEN! Objects = %d\n", i);
  //    exit(-1);
  //  } 
  r = nearestS;
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
  }
  nearest = distances[0];
  for (i =0; i< OBJECTS; i++) {
    if (distances[i] < nearest) {
      nearest = distances[i];
      d2 = i;
    }
  }
  if (d1 ==d2) return 1;
  return 0;
}

/**************
 ** calculate tapestry distance in locating objects
 **/
int hop_calc_loc_TAP(Graph *g, Vertex *client, Vertex *root, Vertex *server,
		      int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
		      int TAPESTRY_NAMESPACE_BASE) {
  int j, distance, mask, r_id, digit, d;
  Vertex *v1, *v2, *r;
  Vertex **routing_table;

  v1 = client;
  r = root;
  distance = 0;
  
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
    if (v1->picked_m == 1) {
      break;
    }
  }

  dijkstra(v1, server, g, NULL);
  d = distance + server->dist;

  r = server;
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
  }

  printf("%d     %d     ",d, distance);
  return distance;
}

/**************
 ** calculate pastry distance in locating objects
 **/
int hop_calc_loc_PAT(Graph *g, Vertex *client, Vertex *root, Vertex *server,
		     int TAPESTRY_NODEID_SIZE, int TAPESTRY_DIGIT_SIZE,
		     int TAPESTRY_NAMESPACE_BASE) {
  int j, distance, mask, r_id, digit;
  Vertex *v1, *v2, *r;
  Vertex **routing_table;

  v1 = client;
  r = root;
  distance = 0;
  
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
  }

  v1 = root;
  r = server;
  
  for (j=0; j< TAPESTRY_NODEID_SIZE; j++) {
    routing_table = (Vertex **) v1->rtable;
    mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *j);
    tapestry_nodeID_atoi(r->nodeid, &r_id,
			 TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
    digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *j);
    v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j +digit];
    dijkstra(v1, v2, g, NULL);
    distance += v2->dist;
    v1 = v2;
  }
  printf("%d\n", distance);
  return distance;
}

/************************************************/
/* calculate shortest path distances from       */
/* the sender to each receiver in the           */
/* recvr_array and write to hop_outfile         */
/************************************************/

void hop_calc_sp(Graph *g, Vertex *sender, Vertex **recvr_array, int n,
                                                        char *hop_outfile) {
  int i, j;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v = NULL;

  printf("\nhop_calc_sp:  sender %s (%s)\n", sender->name, sender->nodeid);
  for (i = 0; i < n; i++) {
    dijkstra(sender, recvr_array[i], g, NULL);
    hop_array[i] = (recvr_array[i])->dist;
    printf("%s(%ld): ", (recvr_array[i])->nodeid, hop_array[i]);

    v = recvr_array[i];
    for (j = 0; j < hop_array[i]; j++) {
      printf("%s ", v->backlink->nodeid);
      v = v->backlink;
    }
    printf("\n");
  }

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
} 
  
/************************************************/
/* calculate number of hops in tapestry routing */
/* from the sender to each receiver in the      */
/* recvr_array and write to the hop_outfile     */
/************************************************/
   
void hop_calc_tapestry(Graph *g, Vertex *sender, Vertex **recvr_array, int n,
                char *hop_outfile, int TAPESTRY_NODEID_SIZE,
                int TAPESTRY_DIGIT_SIZE, int TAPESTRY_NAMESPACE_BASE) {

  int i, j, distance, mask, digit, r_id;
  long *hop_array = (long *) malloc(sizeof(long) * n);
  FILE *hop_fp = fopen(hop_outfile, "w");
  Vertex *v1, *v2, *r;
  Vertex **routing_table;

  printf("\nhop_calc_tapestry:  sender %s (%s)\n",
					sender->name, sender->nodeid);

  for (i = 0; i < n; i++) {
    v1 = sender;
    r = recvr_array[i];
    distance = 0;
    for (j = 0; j < TAPESTRY_NODEID_SIZE; j++) {
      routing_table = (Vertex **) v1->rtable;
      mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE * j);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
                        TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE * j);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE * j + digit];
      dijkstra(v1, v2, g, NULL);
      printf("%s(%ld) ", v2->nodeid, v2->dist);
      distance += v2->dist;
      v1 = v2;
    }
    hop_array[i] = distance;
    printf(" %s(%ld)\n", (recvr_array[i])->nodeid, hop_array[i]);
  }
  printf("\n");

  // write shortest paths to file in binary
  fwrite(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/* read n shortest path distances from a file   */
/************************************************/
/*
void hop_read(long *hop_array, int n, char *hop_outfile) {
  FILE *hop_fp = fopen(hop_outfile, "r");

  fread(hop_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
  }*/

/************************************************/
/* read n neighbors from a file			*/
/************************************************/

void hop_read(long *nbr_array, int n, char *hop_outfile) {
  FILE *hop_fp = fopen(hop_outfile, "r");

  fseek(hop_fp, sizeof(long) * n, SEEK_SET);
  fread(nbr_array, sizeof(long), n, hop_fp);
  fclose(hop_fp);
}

/************************************************/
/* calculate RDP between two members along the  */
/* tapestry overlay to the unicast delay        */
/* between them                                 */
/************************************************/

void rdp_calc(long *hop_array1, long *hop_array2, int n, char *rdp_outfile) {
  int i;
  FILE *rdp_fp = fopen(rdp_outfile, "w");
  float *rdp_array = (float *) malloc(sizeof(float) * n);

  printf("rdp_calc:\n");
  for (i = 0; i < n; i++) {
    rdp_array[i] = (float) (hop_array2[i] * 1.0 / hop_array1[i]);
    printf("%.2f ", rdp_array[i]);
  }
  printf("\n");

  // write relative delay penalties to file in binary
  fwrite(rdp_array, sizeof(float), n, rdp_fp);
  fclose(rdp_fp);
}
