#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <gb_flip.h>
#include <gb_graph.h>
#include <gb_save.h>
#include <gb_dijk.h>

#include "exp.h"
#include "pick.h"
#include "graph.h"
#include "hop.h"
#include "util.h"
#include "nodeid.h"

/************************************************/
/* experiment 3 & 8				*/
/* varying multicast group size			*/
/************************************************/

void exp_3_8() {
  char *topology_fn_array[9] = {"ts5000.gb",
	"ts10000.gb", "ts20000.gb", "ts30000.gb", "ts40000.gb", "ts50000.gb", 
	"ti5000.gb", "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out3_8";
  int i, group_size;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 6;		// 6
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 4096;	// 4096

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 1; i < 5; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
		TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      for (group_size = 16; group_size <= 4096; group_size *= 2) {
        exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
		group_size, vnn);
      }
    }
  }
}

/************************************************/
/* experiment 5 & 10                            */
/* varying tapestry namespace size		*/
/************************************************/

void exp_5_10() {
  char *topology_fn_array[4] = {"ts5000.gb", "ti5000.gb",
                                "AS010200.gb", "mb022399.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out5_10";
  int i, group_size = 64;
  int vnn;
  
  int TAPESTRY_NODEID_SIZE;		// 3, 4, 5, 6 
  int TAPESTRY_DIGIT_SIZE = 2;		// quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE;		// 64, 256, 1024, 4096 

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 4; i++) {
      for (TAPESTRY_NODEID_SIZE = 3; TAPESTRY_NODEID_SIZE <= 6; 
						TAPESTRY_NODEID_SIZE++) {
        TAPESTRY_NAMESPACE_SIZE = pow(TAPESTRY_NAMESPACE_BASE, 
					TAPESTRY_NODEID_SIZE);
        g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
	exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
	
      }
    }
  }
}

/************************************************/
/* experiment 4 & 9				*/
/* varying topology size			*/
/************************************************/

void exp_4_9() {
  char *topology_fn_array[14] = {"ts100.gb", "ts500.gb", "ts1000.gb",
		"ts2000.gb", "ts3000.gb", "ts4000.gb", "ts5000.gb"};
  char out_dir[MAX_FILENAME_LEN] = "out4_9";
  int i, group_size = 64;
  int vnn;

  int TAPESTRY_NODEID_SIZE = 3;
  int TAPESTRY_DIGIT_SIZE = 2;          // quad digit
  int TAPESTRY_NAMESPACE_BASE = 4;
  int TAPESTRY_NAMESPACE_SIZE = 64;

  Graph *g;

  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
    for (i = 0; i < 7; i++) {
      g = exp_init(topology_fn_array[i], TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
      exp_run(g, topology_fn_array[i], out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
    }
  }
}

/************************************************/
/* experiment 11				*/
/* varying tapestry namespace base (fanout)	*/
/************************************************/

void exp_11() {
  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  int group_size = 64;
  int vnn;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
	TAPESTRY_NAMESPACE_SIZE;
  Graph *g;

//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  for (vnn = 5; vnn < 10; vnn++) {
    TAPESTRY_NODEID_SIZE = 3; 
    TAPESTRY_DIGIT_SIZE = 4;
    TAPESTRY_NAMESPACE_BASE = 16;	// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 4; 
    TAPESTRY_DIGIT_SIZE = 3;
    TAPESTRY_NAMESPACE_BASE = 8;		// octal digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 6; 
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;		// quad digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);

    TAPESTRY_NODEID_SIZE = 12; 
    TAPESTRY_DIGIT_SIZE = 1;
    TAPESTRY_NAMESPACE_BASE = 2;		// binary digit
    TAPESTRY_NAMESPACE_SIZE = 4096;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    exp_run(g, topology_fn, out_dir,
                TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE,
                group_size, vnn);
  }
}

/**
 * experiment doing archival reads of objects given multiple roots
 * and given some probability of server load
 */
void exp_archival(int switchy, int hiloload) {
  char topology_fn[MAX_FILENAME_LEN] = "ts5000.gb";
  int mask, mask2, r_id, digit, i, j, k, z, RUNS=100;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int OBJECTS = 32;
  int THRESHOLD = OBJECTS / 2;
  int MAX_EXPERIMENTS = 1;
  int NUMBER_LOADED_SERVERS = 16;
  int LOAD_LATENCY_LOW = 1;
  int LOAD_LATENCY_HIGH = 6;
  float FAILRATE;

  Vertex *client, *root, *v1, *v2, *r, **routing_table;
  Vertex ** servers;
  int serverload[OBJECTS];

  if (hiloload == 1) {
    LOAD_LATENCY_LOW = 5;
    LOAD_LATENCY_HIGH = 10;
  }

  servers = (Vertex **) malloc(sizeof(Vertex *) *OBJECTS);
  
  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;

  for (k = 0; k < MAX_EXPERIMENTS; k++) {
    g = exp_init (topology_fn, TAPESTRY_NODEID_SIZE,
		  TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    clear_picked_m_field(g);
    root = pick_0(g);
    
    r = root;
    for (j=0; j<OBJECTS; j++) {
      //for loop to publish each object and mark nodes along path
      servers[j] = pick_0(g);
      mask2 = 1 << j;
      root->picked_m = root->picked_m | mask2;
      servers[j]->picked_m = servers[j]->picked_m | mask2;
      if (j<NUMBER_LOADED_SERVERS) 
	serverload[j] = LOAD_LATENCY_LOW;
      else serverload[j] = LOAD_LATENCY_HIGH;
      
      v1 = servers[j];
      for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
	tapestry_nodeID_atoi(r->nodeid, &r_id, 
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE *i +digit];
	v2->picked_m = v2->picked_m |mask2;
	v1= v2;
      }
    }
    

    if (switchy == 13) {
      printf("16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31\n");
      for (z=0; z<300; z++) {
	client = pick_0(g);
	printf("%d , ", hop_calc_loc_IP(g, client, root));
	for (j = 0; j<RUNS; j++) {
	  for (i=THRESHOLD; i<OBJECTS; i++) {  //VARYING # OF REQUESTS
	    hop_calc_archival(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			      TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			      OBJECTS, THRESHOLD, i, serverload, 0.0);
	  }
	  printf("\n");
	}
      }
    } else if (switchy == 14) {
      client = pick_0(g);
      for (FAILRATE=0.00; FAILRATE < 0.031; FAILRATE=FAILRATE+0.005) {
	printf("--------------FailRate:%f---------------------\n",
	       FAILRATE);
	for (j = 0; j<RUNS; j++) {
	  for (i=THRESHOLD; i<OBJECTS; i++) {  //VARYING # OF REQUESTS
	    hop_calc_archival(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			      TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			      OBJECTS, THRESHOLD, i, serverload,
			      FAILRATE);
	  }
	  printf("\n");
	}
      }
    } // end else if
  }
}

void exp_multobjLocality() {
  char topology_fn[MAX_FILENAME_LEN] = "mb022399.gb";
  //  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  //  int group_size = 64;
  int vnn, mask, mask2, r_id, digit, i, j, RUNS=200;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int OBJECTS = 5;
  Vertex *client, *root, *v1, *v2, *r, **routing_table;
  Vertex ** servers;
 
  servers = (Vertex **) malloc(sizeof(Vertex *) * OBJECTS);
//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  for (vnn = 0; vnn < 5; vnn++) {
    TAPESTRY_NODEID_SIZE = 4;
    TAPESTRY_DIGIT_SIZE = 2;
    TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
    TAPESTRY_NAMESPACE_SIZE = 256;
    g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
                TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

    clear_picked_m_field(g);
    root = pick_0(g);
    root->picked_m = 1;
    r = root;

    for (j=0; j<OBJECTS; j++) {
      servers[j] = pick_0(g);
      mask2 = 1 << j;
      root->picked_m = root->picked_m | mask2;
      servers[j]->picked_m = servers[j]->picked_m | mask2;
      // mark the nodes from server up to object root
      v1 = servers[j];
      for (i=0; i< TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	
	mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i +digit];
	v2->picked_m = v2->picked_m | mask2;
	v1 = v2;
      }
    }

    printf("Run #%d: \n", vnn);
    for (i=0; i< RUNS; i++) {
      client = pick_0(g);
      printf("%f\n", 
	     hop_calc_loc_objs(g, client, root, servers, TAPESTRY_NODEID_SIZE,
			       TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
			       OBJECTS));
    }
  }
}

void exp_multrootLocality() {
  char topology_fn[MAX_FILENAME_LEN] = "ts5000.gb";
  int vnn, mask, mask2, r_id, digit, i, j, RUNS=200;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int PLANES = 5;
  Vertex *client, *server, *v1, *v2, *r, **routing_table;
  Vertex ** roots;
 
  roots = (Vertex **) malloc(sizeof(Vertex *) * PLANES);
//  for (vnn = 0; vnn < MAX_NUM_EXPERIMENT; vnn++) {
  TAPESTRY_NODEID_SIZE = 6;
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
  g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
	       TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

  for (vnn = 0; vnn < 5; vnn++) {
    clear_picked_m_field(g);
    server = pick_0(g);

    printf("Run #%d: \n", vnn);

    for (j=0; j<PLANES; j++) {
      roots[j] = pick_0(g);
      mask2 = 1 << j;
      roots[j]->picked_m = roots[j]->picked_m | mask2;
      server->picked_m = server->picked_m | mask2;
      // mark the nodes from server up to object root
      v1 = server;
      r = roots[j];
      for (i=0; i< TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	
	mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i +digit];
	v2->picked_m = v2->picked_m | mask2;
	v1 = v2;
      }
    }

    for (i=0; i< RUNS; i++) {
      client = pick_0(g);
      hop_calc_loc_IP_print(g, client, server);
      hop_calc_loc_planes(g, client, roots, server, 
			  TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE,
			  TAPESTRY_NAMESPACE_BASE, PLANES);
      printf("\n");
    }
  }
}

void exp_oneobjLocality() {
  char topology_fn[MAX_FILENAME_LEN] = "ts5000.gb";
  //  char out_dir[MAX_FILENAME_LEN] = "out11_64";
  //  int group_size = 64;
  int vnn, mask, r_id, digit, i, j, k, far;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  int MAXNUM_EXPERIMENT = 5;
  int MAXTRIES = 100;
  Graph *g;

  Vertex *client, *server, *root, *v1, *v2, *r, **routing_table;

  TAPESTRY_NODEID_SIZE = 6; 
  TAPESTRY_DIGIT_SIZE = 2;
  TAPESTRY_NAMESPACE_BASE = 4;	// hexadecimal digit
  TAPESTRY_NAMESPACE_SIZE = 4096;
  g = exp_init(topology_fn, TAPESTRY_NODEID_SIZE,
	       TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);

  for (vnn = 0; vnn < MAXNUM_EXPERIMENT; vnn++) {
    server = pick_0(g);
    root = pick_0(g);
    //here we assume max distance between two hops is around 15
    if ((far = hop_calc_loc_IP(g, server, root) / 15.0) 
	> (vnn+1.0)/MAXNUM_EXPERIMENT
	|| (far < ((float) vnn) / MAXNUM_EXPERIMENT)) {
      vnn--;
      continue;
    }
    
    clear_picked_m_field(g);
    server->picked_m = 1;
    root->picked_m = 1;

    v1 = server;
    r = root;
    
    printf("Run # %d, Server to root distance = %d\n", vnn, far);
    // mark the nodes from server up to object root
    for (i=0; i< TAPESTRY_NODEID_SIZE; i++) {
      routing_table = (Vertex **) v1->rtable;

      mask = (TAPESTRY_NAMESPACE_BASE -1) << (TAPESTRY_DIGIT_SIZE *i);
      tapestry_nodeID_atoi(r->nodeid, &r_id,
			   TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
      digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
      v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i +digit];
      v2->picked_m = 1;
      v1 = v2;
    }
    
    printf("Run #%d: \n", vnn);
    for (i=3; i< 14; i++) {
      for (j=0; j<20; j++) {
	for (k=0; k<MAXTRIES; k++) {
	  client = pick_0(g);
	  far = hop_calc_loc_IP(g, client, server);
	  if (far == i) break;
	}
	printf("Distance= %d ,",far);
	hop_calc_loc_TAP(g, client, root, server, TAPESTRY_NODEID_SIZE,
			 TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
	hop_calc_loc_PAT(g, client, root, server, TAPESTRY_NODEID_SIZE,
			 TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE);
	printf("\n");	
      }
    }
  }
}

/**
 * experiment testing the efficiency (RDP) of routing messages from
 * a normal (Tapestry) node to a mobile Tapestry node using the backtracking
 * pointers.
 */
void exp_mobile_routing_rdp(char * datafile) {

  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  int mask, mask2, r_id, digit, i, j, k, z, distance, RUNS = 100;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int ROOTS = 5;
  int MAX_EXPERIMENTS=3;
  int x;

  Vertex *mclient, *mhost, *v1, *v2, *v3, *r, **routing_table;
  Vertex **rootids;

  rootids = (Vertex **) malloc (sizeof(Vertex *) * ROOTS);

  TAPESTRY_NODEID_SIZE = 6; //# of digits in nodeid
  TAPESTRY_DIGIT_SIZE = 2;  //# of bits in digit, base 4 = 2 bits
  TAPESTRY_NAMESPACE_BASE = 4; //actual base of the digit
  TAPESTRY_NAMESPACE_SIZE = 4096; //# of Tapestry nodes

  for (k = 0; k< MAX_EXPERIMENTS; k++) {
    if (datafile == NULL) {
      printf("STARTING MOBILE TEST with %s.\n", topology_fn);
      g = exp_init (topology_fn, TAPESTRY_NODEID_SIZE,
		    TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    } else {
      printf("STARTING MOBILE TEST with %s.\n", datafile);
      g = exp_init (datafile, TAPESTRY_NODEID_SIZE,
		    TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    }

    //    reset_arc_len(g);

    clear_picked_m_field(g); //clearing marked nodes
    mhost = pick_0(g);  //picking node without marking it
    
    for (j=0; j<ROOTS; j++) {
      //let's publish the same mobilenode on multiple root names
      rootids[j] = pick_0(g);
      mask2 = 1 << j;   // used to mark nodes for a certain object
      mhost->picked_m = mhost->picked_m |mask2;
      rootids[j]->picked_m = rootids[j]->picked_m |mask2;

      distance = 0;

      r = rootids[j];
      v1 = mhost;
      for (i = 0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE*i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i + digit];
	dijkstra(v1, v2, g, NULL);
	//	distance += v2->dist;
	//	if (v2->tapdist != 0 && v2->tapdist > distance)
	// v2->tapdist = distance;  //storing dist from mh in "tapdist" integer
	v2->picked_m = v2->picked_m | mask2;
	v1 = v2;
      }
    }  // done publishing from mobile node to multiple roots

    printf("Run #%d: \n", k);
    for (z=0; z<RUNS; z++) {  //iterate on # of runs, w/ diff mclients
      for (x=0; x<100; x++) {
	/* try 100 times to get something within 300 ms of mclient */
	mclient = pick_0(g);
	if ((j = hop_calc_loc_IP(g, mclient, mhost)) < 300) {
	  printf(" %d ", j);
	  break;
	}
      }
      printf("Over-dist: ");
      for (j=0; j< ROOTS; j++) {
	distance = 0;
	v1 = mclient;
	r = rootids[j];
	for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	  routing_table = (Vertex ** ) v1->rtable;
	  
	  mask = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*i);
	  tapestry_nodeID_atoi(r->nodeid, &r_id,
			       TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	  digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE*i);
	  v2 = routing_table[TAPESTRY_NAMESPACE_BASE *i + digit];
	  dijkstra(v1, v2, g, NULL);
	  distance += v2->dist;
	  //might need to differentiate between diff roots here??
	  if (v2->picked_m != 0) {
	    //	    printf("@%d: %d->", i, distance);
	    break;
	  }
	  v1 = v2;
	}
	//now we have the intersection node V2
	//add on distance from mobile node
	v1 = mhost;
	for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	  routing_table = (Vertex ** ) v1->rtable;
	  
	  mask = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*i);
	  tapestry_nodeID_atoi(r->nodeid, &r_id,
			       TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	  digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE*i);
	  v3 = routing_table[TAPESTRY_NAMESPACE_BASE *i + digit];
	  dijkstra(v1, v3, g, NULL);
	  distance += v3->dist;
	  //might need to differentiate between diff roots here??
	  if (v3 == v2) {
	    //we're back at the intersection
	    break;
	  }
	  v1 = v3;
	}
	printf("%d  ", distance);
      }

      //now we add on some mobileIP distances
      printf("|  ");
      for (i=0; i<200; i++) {
        v1 = pick_0(g);
        dijkstra(mclient, v1, g, NULL);
        distance = v1->dist;
        dijkstra(v1, mhost, g, NULL);
        printf("%d->", distance);
        distance += mhost->dist;
        printf("%d  ", distance);
      }
      printf("\n"); //end of single run for all the roots
    }
  }
}


/**
 * experiment testing the load placed on nodes/routers by update messages
 * in a system where a lot of nodes are migrating. showing that HA is 
 * a pt of centralization that does not scale.
 */
void exp_mobile_routing_load(char * datafile) {

  char topology_fn[MAX_FILENAME_LEN] = "ti5000.gb";
  int mask, r_id, digit, i, k, z, RUNS = 1000;
  int TAPESTRY_NODEID_SIZE, TAPESTRY_DIGIT_SIZE, TAPESTRY_NAMESPACE_BASE,
    TAPESTRY_NAMESPACE_SIZE;
  Graph *g;
  int MAX_EXPERIMENTS=10;

  Vertex *mclient, *mhost, *v1, *v2, *v3, *r, **routing_table, *ha;
  Vertex **rootids;

  rootids = (Vertex **) malloc (sizeof(Vertex *) * 1);

  TAPESTRY_NODEID_SIZE = 6; //# of digits in nodeid
  TAPESTRY_DIGIT_SIZE = 2;  //# of bits in digit, base 4 = 2 bits
  TAPESTRY_NAMESPACE_BASE = 4; //actual base of the digit
  TAPESTRY_NAMESPACE_SIZE = 4096; //# of Tapestry nodes

    if (datafile == NULL) {
      printf("STARTING MOBILE TEST with %s.\n", topology_fn);
      g = exp_init (topology_fn, TAPESTRY_NODEID_SIZE,
		    TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    } else {
      printf("STARTING MOBILE TEST with %s.\n", datafile);
      g = exp_init (datafile, TAPESTRY_NODEID_SIZE,
		    TAPESTRY_NAMESPACE_BASE, TAPESTRY_NAMESPACE_SIZE);
    }
    reset_arc_len(g);

  for (k = 0; k< MAX_EXPERIMENTS; k++) {
    printf("Run #%d: \n", k);
    ha = pick_0(g);
    r = ha;
    clear_load_fields(g);
    for (z=1; z<=RUNS; z++) {  //iterate on # of runs, w/ diff mclients
      mclient = pick_0(g);
      mhost = pick_0(g);  //picking 
      clear_picked_m_field(g); //clearing marked nodes

      mhost->picked_m = 1;   //begin publishing
      r->picked_m = 1;
      v1 = mhost;
      for (i = 0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex **) v1->rtable;
	mask = (TAPESTRY_NAMESPACE_BASE - 1) << (TAPESTRY_DIGIT_SIZE*i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id &mask) >> (TAPESTRY_DIGIT_SIZE *i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE * i + digit];
	v2->picked_m = 1;
	v1 = v2;
      }                      //finished publishing
      
      v1 = mclient;
      for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex ** ) v1->rtable;
	mask = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE*i);
	v2 = routing_table[TAPESTRY_NAMESPACE_BASE *i + digit];
	dijkstra(v1, v2, g, NULL);
	mark_load_field_tap(g, v1, v2);
	if (v2->picked_m != 0) {
	  //	    printf("@%d: %d->", i, distance);
	  break;
	}
	v1 = v2;
      }
      //now we have the intersection node V2
      //add on distance from mobile node
      v1 = mhost;
      for (i=0; i<TAPESTRY_NODEID_SIZE; i++) {
	routing_table = (Vertex ** ) v1->rtable;
	
	mask = (TAPESTRY_NAMESPACE_BASE-1) << (TAPESTRY_DIGIT_SIZE*i);
	tapestry_nodeID_atoi(r->nodeid, &r_id,
			     TAPESTRY_NODEID_SIZE, TAPESTRY_NAMESPACE_BASE);
	digit = (r_id & mask) >> (TAPESTRY_DIGIT_SIZE*i);
	v3 = routing_table[TAPESTRY_NAMESPACE_BASE *i + digit];
	dijkstra(v1, v3, g, NULL);
	mark_load_field_tap(g, v1, v3);
	if (v3 == v2) {
	  //we're back at the intersection
	  break;
	}
	v1 = v3;
      }

      //now we add on a mobileIP distances
      mclient = pick_0(g);  //vary our new proxy location
      dijkstra(mclient, ha, g, NULL);
      mark_load_field_mip(g, mclient, ha);
      dijkstra(ha, mhost, g, NULL);
      mark_load_field_mip(g, ha, mhost);

      if (z % 100 == 0) {
	r = pick_0(g);
	printf("Run#: %d MIP_load: %d  TAP_LOAD:%d\n", z, 
	       find_max_load_mip(g), find_max_load_tap(g));
      }
    }
  }
}

/************************************************/
/* MAIN						*/
/************************************************/

int main(int argc, char *argv[]) {

  //  int i;
  //  float random;

  gb_init_rand(time(NULL));     // initialize the random number generator

  // experiment for 3 & 8
  // exp_3_8();

  // experiment for 5 & 10
  // exp_5_10();

  // experiment for 4 & 9
  // exp_4_9();

  // experiment for 11
  //  exp_archival(14, 0);       //exp_archival(exp#, load: 1=high, 0=low)
  //  exp_oneobjLocality();
  //   exp_multrootLocality();

  if (argc == 1) 
    exp_mobile_routing_rdp(NULL);
  else 
    exp_mobile_routing_rdp(argv[1]);
  /*
  if (argc == 1)
    exp_mobile_routing_load(NULL);
  else 
    exp_mobile_routing_load(argv[1]);
  */
  
  /*  for (i=0; i<10000; i++) {
    random = ((float) rand()) / RAND_MAX;
    printf("%f\n", -1* log(random));
  }
  */
  return 0;
}
