#include <gb_graph.h>
#include <gb_flip.h>
#include <gb_dijk.h>

#include "graph.h"
#include "pick.h"

/************************************************/
/* clear the picked_m field of each vertex of   */
/* the Graph pointed to by g                    */
/************************************************/

void clear_picked_m_field(Graph *g) {
  Vertex *v;

  for (v = g->vertices; v < g->vertices + g->n; v++) {
    v->picked_m = 0;
  }
}

/************************************************
 * clear the load field of each IP link         *
 ************************************************/

void clear_load_fields(Graph *g) {
  Vertex *v;
  for (v = g->vertices; v<g->vertices+g->n; v++) {
    v->mipload = 0;
    v->tapload = 0;
  }
}

int find_max_load_mip(Graph *g) {
  Vertex *v;
  int max = 0;
  for (v = g->vertices; v < g->vertices+ g->n; v++) {
    if (v->mipload > max)
      max = v->mipload;
  }
  return max;
}

int find_max_load_tap(Graph *g) {
  Vertex *v;
  int max = 0;
  for (v = g->vertices; v < g->vertices+ g->n; v++) {
    if (v->tapload > max)
      max = v->tapload;
  }
  return max;
}

void mark_load_field_mip(Graph *g, Vertex *from, Vertex *to) {
  Vertex *temp;
  for (temp = to; temp != from; temp = temp->backlink) {
    //    if (temp->backlink == NULL) 
    //      fprintf(stderr, "MAJOR ERROR: NODE not seen\n");
    temp->mipload = temp->mipload+1;
  }
}

void mark_load_field_tap(Graph *g, Vertex *from, Vertex *to) {
  Vertex *temp;
  for (temp = to; temp != from; temp = temp->backlink) {
    //    if (temp->backlink == NULL) 
    //      fprintf(stderr, "MAJOR ERROR: NODE not seen\n");
    temp->tapload = temp->tapload+1;
  }
}

/************************************************/
/* pick 1 vertex from the tapestry network      */
/************************************************/

Vertex *pick_1(Graph *g) {
  Vertex *v;

  while (1) {
    v = g->vertices + gb_unif_rand(g->n);
    if (v->picked_t == 1 && v->picked_m == 0) {
      v->picked_m = 1;
      return v;
    }
  }
}

Vertex *pick_0(Graph *g) {
  Vertex *v;

  while (1) {
    v = g->vertices + gb_unif_rand(g->n);
    if (v->picked_t == 1) {
      return v;
    }
  }
}

/************************************************/
/* pick n vertices from the tapestry network    */
/************************************************/

void pick_n(Graph *g, Vertex **vertex_array, int n) {
  Vertex *v;
  int i;

  for (i = 0; i < n; ) {
    v = g->vertices + gb_unif_rand(g->n);
    if (v->picked_t == 1 && v->picked_m == 0) {
      v->picked_m = 1;
      vertex_array[i++] = v;
    }
  }
}
